/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * packet.c
 * --------
 *
 * Initial Version	Thu May 16 17:36:09 CEST 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <features.h>
#if __GLIBC__ >=2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#endif
#include <errno.h>

#include "packet.h"
#include "debug.h"
#include "dhcpd.h"
#include "options.h"
#include "arpping.h"



/* ************************************************************************* */
/* *** TYPEDEFS ************************************************************ */

struct dhcp_packet {
	unsigned char		pad[2];		/* to align ip header */
	struct ethhdr 		ethhdr;
	struct iphdr		ip;
	struct udphdr		udp;
	struct dhcpMessage	data;
} __attribute__ ((packed));



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

/* read a packet from socket fd */
int
get_packet( struct dhcpMessage *packet,int fd )
{
	static char	*_f = "get_packet:";
	int	bytes;
	int	i;
	char	broken_vendors[][8] = {
		"MSFT 98",
		""
	};
	char	*vendor;

	memset( packet,0,sizeof( struct dhcpMessage ));
	bytes = read( fd,packet,sizeof( struct dhcpMessage ));
	if (bytes < 0) {
		ERR( "%s read failed errno=%s\n",_f,strerror( errno ));
		return( -1 );
	}

	if (ntohl(packet->cookie) != DHCP_MAGIC) {
		LOG( "received bogus message -- ignoring\n" );
		return( -1 );
	}
	
	if (packet->op == BOOTREQUEST && (vendor = get_option( packet,DHCP_VENDOR ))) {
		for (i = 0; broken_vendors[i][0]; i++) {
			if (vendor[OPT_LEN - 2] == (signed) strlen(broken_vendors[i]) &&
			    !strncmp(vendor, broken_vendors[i], vendor[OPT_LEN - 2]) &&
			    !(ntohs(packet->flags) & BROADCAST_FLAG)) {
			    	DEB( "%s broken client (%s), forcing broadcast\n",_f,broken_vendors[i] );
			    	packet->flags |= htons(BROADCAST_FLAG);
			}
		}
	}
	return( bytes );
}

/* Calculate the length of a packet, and make sure its a multiple of 2 */
static int
calc_length( struct dhcpMessage *payload )
{
	int	payload_length;

	payload_length = sizeof( struct dhcpMessage ) - 308;
	payload_length += end_option( payload->options ) + 1;
	if (payload_length % 2) {
		payload_length++;
		*((char *) payload + payload_length - 1) = '\0';
	}
	DEB( "payload length is %d bytes\n",payload_length );
	return( payload_length );
}


u_int16_t
checksum( void *addr,int count )
{
	/*
	 * Compute Internet Checksum for "count" bytes
	 * beginning at location "addr".
	 */
	register int32_t	sum = 0;
	u_int16_t		*source = (u_int16_t *) addr;

	while( count > 1 )  {
		/*  This is the inner loop */
		sum += *source++;
		count -= 2;
	}

	/*  Add left-over byte, if any */
	if( count > 0 ) {
		sum += * (unsigned char *) source;
	}

	/*  Fold 32-bit sum to 16 bits */
	while (sum>>16) {
		sum = (sum & 0xffff) + (sum >> 16);
	}

	return( ~sum );
}

/*
 * Constuct a ip/udp header for a packet, and specify the source
 * and dest hardware address
 */
int raw_packet( struct dhcpMessage *payload,
		u_int32_t source_ip,
		int source_port,
		char *source_arp,
		u_int32_t dest_ip,
		int dest_port,
		char *dest_arp,
		char *interface )
{
	static char		*_f = "raw_packet:";
	int			fd;
	int			result;
	int			payload_len;
	int			total_len;
	struct sockaddr		addr;
	struct dhcp_packet	packet;

	if (openRawSocket( &fd,ETH_P_IP ) < 0) {
		return( -1 );
	}
	memset( &addr,0,sizeof( addr ));
	memset( &packet,0,sizeof( packet ));
	strcpy( addr.sa_data,interface );
	if (bind( fd,(struct sockaddr *)&addr,sizeof(addr)) < 0) {
		ERR( "%s bind failed errno=%s\n",_f,strerror( errno ));
		close( fd );
		return( -1 );
	}

	memcpy( packet.ethhdr.h_dest,dest_arp,ETH_ALEN );
	memcpy( packet.ethhdr.h_source,source_arp,ETH_ALEN );
	packet.ethhdr.h_proto = htons( ETH_P_IP );

	payload_len        = calc_length( payload );
	total_len          = sizeof(packet.ip)+sizeof(packet.udp)+payload_len;
	packet.ip.protocol = IPPROTO_UDP;
	packet.ip.saddr    = source_ip;
	packet.ip.daddr    = dest_ip;
	/* cheat on the psuedo-header */
	packet.ip.tot_len  = htons(sizeof(packet.udp) + payload_len);
	packet.udp.source  = htons(source_port);
	packet.udp.dest    = htons(dest_port);
	packet.udp.len     = htons(sizeof(packet.udp) + payload_len);
	memcpy( &(packet.data),payload,payload_len );
	packet.udp.check   = checksum( &(packet.ip),total_len );
	packet.ip.tot_len  = htons( total_len );
	packet.ip.ihl      = sizeof(packet.ip) >> 2;
	packet.ip.version  = IPVERSION;
	packet.ip.ttl      = IPDEFTTL;
	packet.ip.check    = checksum( &(packet.ip),sizeof( packet.ip ));

	memset( &addr,0,sizeof( addr ));
	strcpy( addr.sa_data,interface );
	result = sendto( fd,
			 &(packet.ethhdr),
			 total_len + sizeof( packet.ethhdr ),
			 0,
			 (struct sockaddr *)&addr,
			 sizeof(addr) );

	if (result <= 0) {
		ERR( "%s sendto failed errno=%s\n",_f,strerror( errno ));
	}

	close( fd );
	return( result );
}


/* Let the kernel do all the work for packet generation */
int
kernel_packet( struct dhcpMessage *payload,
               u_int32_t source_ip,
               int source_port,
               u_int32_t dest_ip,
               int dest_port )
{
	static char		*_f = "kernel_packet:";
	int			n = 1;
	int			fd;
	int			result;
	int			payload_length = calc_length(payload);
	struct sockaddr_in	client;
	
	if ((fd = socket( PF_INET,SOCK_DGRAM,IPPROTO_UDP )) < 0) {
		ERR( "%s socket failed errno=%s\n",_f,strerror( errno ));
		return( -1 );
	}
	
	if (setsockopt( fd,SOL_SOCKET,SO_REUSEADDR,(char *)&n,sizeof(n)) == -1) {
		ERR( "%s setsockopt(SO_REUSEADDR) failed errno=%s\n",
		      _f,strerror( errno ));
		close( fd );
		return( -1 );
	}

	memset( &client,0,sizeof( client ));
	client.sin_family = AF_INET;
	client.sin_port = htons( source_port );
	client.sin_addr.s_addr = source_ip;
	if (bind( fd,(struct sockaddr *)&client,sizeof(struct sockaddr)) == -1) {
		ERR( "%s bind failed errno=%s\n",_f,strerror( errno ));
		close( fd );
		return( -1 );
	}

	memset(&client, 0, sizeof(client));
	client.sin_family = AF_INET;
	client.sin_port = htons(dest_port);
	client.sin_addr.s_addr = dest_ip; 

	if (connect( fd,(struct sockaddr *)&client,sizeof(struct sockaddr)) == -1) {
		ERR( "%s connect failed errno=%s\n",_f,strerror( errno ));
		close( fd );
		return( -1 );
	}

	result = write( fd,payload,payload_length );
	close( fd );
	return( result );
}
