/*
 * Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH> 
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * serverpacket.c
 * --------------
 *
 * Initial Version	Thu May 16 17:36:09 CEST 2002 - jse
 * Rewritten based on uDhcpd
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

/* ************************************************************************* */
/* *** INCLUDES ************************************************************ */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <time.h>

#include "packet.h"
#include "debug.h"
#include "dhcpd.h"
#include "options.h"
#include "leases.h"



/* ************************************************************************* */
/* *** STATIC FUNCTIONS **************************************************** */

/* send a packet to giaddr using the kernel ip stack */
static int
send_packet_to_relay( struct dhcpMessage *payload )
{
	int	rc;

	DEB( "forwarding packet to relay\n" );
	rc = kernel_packet( payload,
	                    server_config.server,
	                    SERVER_PORT,
	                    payload->giaddr,
	                    SERVER_PORT );
	return( rc );
}

/*
 * send a packet to a specific arp address and ip address
 * by creating our own ip packet
 */
static int
send_packet_to_client( struct dhcpMessage *payload,int force_broadcast )
{
	u_int32_t	ciaddr;
	int		rc;
	char		chaddr[6];
	
	if (force_broadcast) {
		INF( "broadcasting packet to client (NAK)\n" );
		ciaddr = INADDR_BROADCAST;
		memcpy( chaddr,MAC_BCAST_ADDR,6 );
	} else if (payload->ciaddr) {
		INF( "unicasting packet to client ciaddr\n" );
		ciaddr = payload->ciaddr;
		memcpy( chaddr,payload->chaddr,6 );
	} else if (ntohs(payload->flags) & BROADCAST_FLAG) {
		INF( "broadcasting packet to client (requested)\n" );
		ciaddr = INADDR_BROADCAST;
		memcpy( chaddr,MAC_BCAST_ADDR,6 );
	} else {
		INF( "unicasting packet to client yiaddr\n" );
		ciaddr = payload->yiaddr;
		memcpy( chaddr,payload->chaddr,6 );
	}
	rc = raw_packet( payload,
			 server_config.server,SERVER_PORT,server_config.arp,
			 ciaddr,CLIENT_PORT,chaddr,
			 server_config.interface );
	return( rc );
}

/*
 * send a dhcp packet, if force broadcast is set,
 * the packet will be broadcast to the client
 */
static int
send_packet( struct dhcpMessage *payload,int force_broadcast )
{
	int	ret;

	if (payload->giaddr) {
		ret = send_packet_to_relay( payload );
	} else {
		ret = send_packet_to_client( payload,force_broadcast );
	}
	return( ret );
}

static void
init_packet( struct dhcpMessage *packet,
             struct dhcpMessage *oldpacket,
             char type )
{
	memset( packet,0,sizeof( struct dhcpMessage ));

	packet->op = BOOTREPLY;
	packet->htype = ETH_10MB;
	packet->hlen = ETH_10MB_LEN;
	packet->xid = oldpacket->xid;
	memcpy(packet->chaddr, oldpacket->chaddr, 16);
	packet->cookie = htonl(DHCP_MAGIC);
	packet->options[0] = DHCP_END;
	packet->flags = oldpacket->flags;
	packet->giaddr = oldpacket->giaddr;
	packet->ciaddr = oldpacket->ciaddr;
	add_simple_option( packet->options,DHCP_MESSAGE_TYPE,type);
	add_simple_option( packet->options,DHCP_SERVER_ID,ntohl( server_config.server ));
}

/* add in the bootp options */
static void
add_bootp_options( struct dhcpMessage *packet )
{
	packet->siaddr = server_config.siaddr;
	if (server_config.sname) {
		strncpy( packet->sname,server_config.sname,sizeof( packet->sname )-1 );
	}
	if (server_config.boot_file) {
		strncpy( packet->file,server_config.boot_file,sizeof( packet->file )-1 );
	}
}



/* ************************************************************************* */
/* *** GLOBAL FUNCTIONS **************************************************** */

/* send a DHCP OFFER to a DHCP DISCOVER */
int
sendOffer( struct dhcpMessage *oldpacket )
{
	struct dhcpMessage	packet;
	struct dhcpOfferedAddr	*lease = NULL;
	u_int32_t		req_align;
	u_int32_t		lease_time_align = server_config.lease;
	char			*req;
	char			*lease_time;
	struct option_set	*curr;
	struct in_addr		addr;

	init_packet( &packet,oldpacket,DHCPOFFER );
	
	/* the client is in our lease/offered table */
	if ((lease = find_lease_by_chaddr( oldpacket->chaddr ))) {
		if (!lease_expired( lease )) {
			lease_time_align = lease->expires - time( 0 );
		}
		packet.yiaddr = lease->yiaddr;

	/* Or the client has a requested ip */
	} else if ((req = get_option( oldpacket,DHCP_REQUESTED_IP )) &&
		   /* Don't look here (ugly hackish thing to do) */
		   memcpy( &req_align,req,4 ) &&

		   /* and the ip is in the lease range */
		   ntohl( req_align ) >= ntohl( server_config.start ) &&
		   ntohl( req_align ) <= ntohl( server_config.end ) &&
		   
		   /* and its not already taken/offered */
		   ((!(lease = find_lease_by_yiaddr( req_align )) ||
		   
		   /* or its taken, but expired */
		   lease_expired( lease ))))
	{
		packet.yiaddr = req_align;
	/* otherwise, find a free IP */
	} else {
		packet.yiaddr = find_address( 0 );
		/* try for an expired lease */
		if (!packet.yiaddr) {
			packet.yiaddr = find_address( 1 );
		}
	}

	if(!packet.yiaddr) {
		LOG( "no IP addresses to give -- OFFER abandoned\n" );
		return( -1 );
	}

	if (!add_lease( packet.chaddr,packet.yiaddr,server_config.offer_time)) {
		LOG( "lease pool is full -- OFFER abandoned\n" );
		return( -1 );
	}

	if ((lease_time = get_option( oldpacket,DHCP_LEASE_TIME ))) {
		memcpy( &lease_time_align,lease_time,4 );
		lease_time_align = ntohl( lease_time_align );
		if (lease_time_align > server_config.lease) {
			lease_time_align = server_config.lease;
		}
	}

	/* Make sure we aren't just using the lease time from the previous offer */
	if (lease_time_align < server_config.min_lease) {
		lease_time_align = server_config.lease;
	}

	add_simple_option( packet.options,DHCP_LEASE_TIME,lease_time_align );

	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME) {
			add_option_string( packet.options,curr->data );
		}
		curr = curr->next;
	}

	add_bootp_options( &packet );
	
	addr.s_addr = packet.yiaddr;
	LOG( "sending OFFER of %s\n",inet_ntoa( addr ));
	return( send_packet( &packet,0 ));
}

int
sendNAK( struct dhcpMessage *oldpacket )
{
	struct dhcpMessage	packet;

	init_packet( &packet,oldpacket,DHCPNAK );
	LOG( "sending NAK\n" );
	return( send_packet( &packet,1 ));
}

int
sendACK( struct dhcpMessage *oldpacket,u_int32_t yiaddr )
{
	struct dhcpMessage	packet;
	struct option_set	*curr;
	char			*lease_time;
	u_int32_t		lease_time_align = server_config.lease;
	struct in_addr		addr;

	init_packet( &packet,oldpacket,DHCPACK );
	packet.yiaddr = yiaddr;

	if ((lease_time = get_option( oldpacket,DHCP_LEASE_TIME ))) {
		memcpy( &lease_time_align,lease_time,4 );
		lease_time_align = ntohl( lease_time_align );
		if (lease_time_align > server_config.lease) {
			lease_time_align = server_config.lease;
		} else if (lease_time_align < server_config.min_lease) {
			lease_time_align = server_config.lease;
		}
	}
	
	add_simple_option( packet.options,DHCP_LEASE_TIME,lease_time_align );
	
	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME) {
			add_option_string( packet.options,curr->data );
		}
		curr = curr->next;
	}

	add_bootp_options( &packet );

	addr.s_addr = packet.yiaddr;
	LOG( "sending ACK to %s\n",inet_ntoa( addr ));

	if (send_packet( &packet,0 ) < 0) {
		return -1;
	}
	add_lease( packet.chaddr,packet.yiaddr,lease_time_align );

	return( 0 );
}


int
send_inform( struct dhcpMessage *oldpacket )
{
	struct dhcpMessage	packet;
	struct option_set	*curr;

	init_packet( &packet,oldpacket,DHCPACK );

	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME) {
			add_option_string( packet.options,curr->data );
		}
		curr = curr->next;
	}

	add_bootp_options( &packet );

	return( send_packet( &packet,0 ));
}
