/***********************************************************************
*
* pppoe.c 
*
* Implementation of user-space PPPoE redirector for Linux.
*
* Copyright (C) 2000-2001 by Roaring Penguin Software Inc.
*
* This program may be distributed according to the terms of the GNU
* General Public License, version 2 or (at your option) any later version.
*
***********************************************************************/


#include "pppoe.h"
#include "debug.h"

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef USE_LINUX_PACKET
#include <sys/ioctl.h>
#include <fcntl.h>
#endif

#include <signal.h>

#ifdef HAVE_N_HDLC
#ifndef N_HDLC
#include <linux/termios.h>
#endif
#endif


/* Global variables -- options */
int optInactivityTimeout = 0;	/* Inactivity timeout */
int optClampMSS          = 0;	/* Clamp MSS to this value */
int optSkipSession       = 0;	/* Perform discovery, print session info
				   and exit */
int optFloodDiscovery    = 0;   /* Flood server with discovery requests.
				   USED FOR STRESS-TESTING ONLY.  DO NOT
				   USE THE -F OPTION AGAINST A REAL ISP */
#ifdef WITH_KERNEL_MODE_PPPOE
#include <net/if_ppp.h>
#include "if_hook.h"
#include "km_pppoe.h"
static int	kernel_session( PPPoEConnection *conn,int clampMss );
static void	kernel_session_finit( PPPoEConnection *conn );
int		optKernelModePPPoE = 0;
char		*optPPPIfname = NULL;
PPPoEConnection	*optPPPoEConn = NULL;
#endif

PPPoEConnection *Connection = NULL; /* Must be global -- used
				       in signal handler */

void doExit( int rc )
{
#ifdef WITH_KERNEL_MODE_PPPOE
	if (optPPPoEConn != NULL && optKernelModePPPoE) {
		kernel_session_finit( optPPPoEConn );
		optPPPoEConn = NULL;
	}
#endif
	LOG( "terminating with status=%s\n",strerror( rc ));
	CLOSELOG();
	exit( rc );
}

/***********************************************************************
*%FUNCTION: sendSessionPacket
*%ARGUMENTS:
* conn -- PPPoE connection
* packet -- the packet to send
* len -- length of data to send
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Transmits a session packet to the peer.
***********************************************************************/
void
sendSessionPacket(PPPoEConnection *conn, PPPoEPacket *packet, int len)
{
    packet->length = htons(len);
    if (optClampMSS) {
	clampMSS(packet, "outgoing", optClampMSS);
    }
    if (sendPacket(conn, conn->sessionSocket, packet, len + HDR_SIZE) < 0) {
	if (errno == ENOBUFS) {
	    /* No buffer space is a transient error */
	    return;
	}
	doExit(EXIT_FAILURE);
    }
#ifdef DUMP_FRAMES
    if (conn->debugFile) {
	dumpPacket(conn->debugFile, packet, "SENT");
	fprintf(conn->debugFile, "\n");
	fflush(conn->debugFile);
    }
#endif
}

#ifdef USE_BPF
/**********************************************************************
*%FUNCTION: sessionDiscoveryPacket
*%ARGUMENTS:
* packet -- the discovery packet that was received
*%RETURNS:
* Nothing
*%DESCRIPTION:
* We got a discovery packet during the session stage.  This most likely
* means a PADT.
*
* The BSD version uses a single socket for both discovery and session
* packets.  When a packet comes in over the wire once we are in
* session mode, either syncReadFromEth() or asyncReadFromEth() will
* have already read the packet and determined it to be a discovery
* packet before passing it here.
***********************************************************************/
void
sessionDiscoveryPacket(PPPoEPacket *packet)
{
    /* Sanity check */
    if (packet->code != CODE_PADT) {
	return;
    }

    /* It's a PADT, all right.  Is it for us? */
    if (packet->session != Connection->session) {
	/* Nope, ignore it */
	return;
    }
    if (memcmp(packet->ethHdr.h_dest, Connection->myEth, ETH_ALEN)) {
	return;
    }

    if (memcmp(packet->ethHdr.h_source, Connection->peerEth, ETH_ALEN)) {
	return;
    }

    syslog(LOG_INFO,
	   "Session terminated -- received PADT from peer");
    parsePacket(packet, parseLogErrs, NULL);
    sendPADT( Connection,"Received PADT from peer" );
    doExit( EXIT_SUCCESS );
}
#else
/**********************************************************************
*%FUNCTION: sessionDiscoveryPacket
*%ARGUMENTS:
* conn -- PPPoE connection
*%RETURNS:
* Nothing
*%DESCRIPTION:
* We got a discovery packet during the session stage.  This most likely
* means a PADT.
***********************************************************************/
void
sessionDiscoveryPacket(PPPoEConnection *conn)
{
    PPPoEPacket packet;
    int len;

    if (receivePacket(conn->discoverySocket, &packet, &len) < 0) {
	return;
    }

    /* Check length */
    if (ntohs(packet.length) + HDR_SIZE > len) {
	ERR( "bogus PPPoE length field (%u)\n",
	       (unsigned int) ntohs(packet.length));
	return;
    }

    if (packet.code != CODE_PADT) {
	/* Not PADT; ignore it */
	return;
    }

    /* It's a PADT, all right.  Is it for us? */
    if (packet.session != conn->session) {
	/* Nope, ignore it */
	return;
    }

    if (memcmp(packet.ethHdr.h_dest, conn->myEth, ETH_ALEN)) {
	return;
    }

    if (memcmp(packet.ethHdr.h_source, conn->peerEth, ETH_ALEN)) {
	return;
    }

#ifdef DUMP_FRAMES
    if (conn->debugFile) {
	dumpPacket(conn->debugFile, &packet, "RCVD");
	fprintf(conn->debugFile, "\n");
	fflush(conn->debugFile);
    }
#endif

    LOG( "session terminated - received PADT from peer\n" );
    parsePacket(&packet, parseLogErrs, NULL);
    sendPADT( Connection,"Received PADT from peer" );
    doExit( EXIT_SUCCESS );
}
#endif /* USE_BPF */

/**********************************************************************
*%FUNCTION: session
*%ARGUMENTS:
* conn -- PPPoE connection info
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Handles the "session" phase of PPPoE
***********************************************************************/
void
session(PPPoEConnection *conn)
{
    fd_set readable;
    PPPoEPacket packet;
    struct timeval tv;
    struct timeval *tvp;
    int maxFD = 0;
    int r;

    /* Open a session socket */
    conn->sessionSocket = openInterface(conn->ifName, Eth_PPPOE_Session, conn->myEth);

    /* Prepare for select() */
    if (conn->sessionSocket > maxFD)   maxFD = conn->sessionSocket;
    if (conn->discoverySocket > maxFD) maxFD = conn->discoverySocket;
    maxFD++;

    /* Fill in the constant fields of the packet to save time */
    memcpy(packet.ethHdr.h_dest, conn->peerEth, ETH_ALEN);
    memcpy(packet.ethHdr.h_source, conn->myEth, ETH_ALEN);
    packet.ethHdr.h_proto = htons(Eth_PPPOE_Session);
    packet.ver = 1;
    packet.type = 1;
    packet.code = CODE_SESS;
    packet.session = conn->session;

    initPPP();

#ifdef USE_BPF
    /* check for buffered session data */
    while (BPF_BUFFER_HAS_DATA) {
	if (conn->synchronous) {
	    syncReadFromEth(conn, conn->sessionSocket, optClampMSS);
	} else {
	    asyncReadFromEth(conn, conn->sessionSocket, optClampMSS);
	}
    }
#endif

    for (;;) {
	if (optInactivityTimeout > 0) {
	    tv.tv_sec = optInactivityTimeout;
	    tv.tv_usec = 0;
	    tvp = &tv;
	} else {
	    tvp = NULL;
	}
	FD_ZERO(&readable);
	FD_SET(0, &readable);     /* ppp packets come from stdin */
	if (conn->discoverySocket >= 0) {
	    FD_SET(conn->discoverySocket, &readable);
	}
	FD_SET(conn->sessionSocket, &readable);
	while(1) {
	    r = select(maxFD, &readable, NULL, NULL, tvp);
	    if (r >= 0 || errno != EINTR) break;
	}
	if (r < 0) {
	    fatalSys( "session: select failed" );
	}
	if (r == 0) { /* Inactivity timeout */
	    ERR( "inactivity timeout\n" );
	    sendPADT( conn,"pppoed: inactivity timeout" );
	    doExit( EXIT_FAILURE );
	}

	/* Handle ready sockets */
	if (FD_ISSET(0, &readable)) {
	    if (conn->synchronous) {
		syncReadFromPPP(conn, &packet);
	    } else {
		asyncReadFromPPP(conn, &packet);
	    }
	}

	if (FD_ISSET(conn->sessionSocket, &readable)) {
	    do {
		if (conn->synchronous) {
		    syncReadFromEth(conn, conn->sessionSocket, optClampMSS);
		} else {
		    asyncReadFromEth(conn, conn->sessionSocket, optClampMSS);
		}
	    } while (BPF_BUFFER_HAS_DATA);
	}

#ifndef USE_BPF	
	/* BSD uses a single socket, see *syncReadFromEth() */
	/* for calls to sessionDiscoveryPacket() */
	if (conn->discoverySocket >= 0) {
	    if (FD_ISSET(conn->discoverySocket, &readable)) {
		sessionDiscoveryPacket(conn);
	    }
	}
#endif

    }
}

#ifdef WITH_KERNEL_MODE_PPPOE
static int
kernel_session( PPPoEConnection *conn,int clampMss )
{
	static char	*_f = "kernel_session:";
	int		fd;
	struct ifreq	ifr;
	struct ifhook	ihk;
	struct km_pppoe	pppoe;
	int		rc;

	if ((fd = socket( PF_INET,SOCK_DGRAM,0 )) < 0) {
		rc = errno;
		ERR( "%s open socket\n",_f );
		return( rc );
	}

	memset ( &pppoe,'\0',sizeof( pppoe ));
	strncpy( pppoe.ifname,conn->ifName,sizeof( pppoe.ifname ));
	memcpy( pppoe.my_eth,conn->myEth,ETH_ALEN );
	memcpy( pppoe.peer_eth,conn->peerEth,ETH_ALEN );
	if (clampMss > 0) {
		pppoe.clamp_mss = clampMss;
	}
	pppoe.session_id = conn->session;

	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,optPPPIfname,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t)&pppoe;
	if (ioctl( fd,SIOCSPPPOECFG,(caddr_t)&ifr ) != 0) {
		rc = errno;
		ERR( "%s ioctl(SIOCSPPPOECFG) failed\n",_f );
		close( fd );
		return( rc );
	}

	memset ( &ihk,'\0',sizeof( ihk ));
	strncpy( ihk.ihk_name,"pppoe-session-recv",sizeof( ihk.ihk_name ));
	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,optPPPIfname,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t) &ihk;
	if (ioctl( fd,SIOCGHOOK,(caddr_t)&ifr ) != 0) {
		rc = errno;
		ERR( "%s ioctl(SIOCGHOOK) failed\n",_f );
		close( fd );
		return( rc );
	}

	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,conn->ifName,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t) &ihk;
	if (ioctl( fd,SIOCSHOOKREGISTER,(caddr_t)&ifr ) != 0) {
		rc = errno;
		ERR( "%s ioctl(SIOCSHOOKREGISTER) failed\n",_f );
		close( fd );
		return( rc );
	}

	memset ( &ihk,'\0',sizeof( ihk ));
	strncpy( ihk.ihk_name,"pppoe-session-xmit",sizeof( ihk.ihk_name ));
	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,optPPPIfname,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t) &ihk;
	if (ioctl( fd,SIOCGHOOK,(caddr_t)&ifr ) != 0) {
		rc = errno;
		ERR( "%s ioctl(SIOCGHOOK) failed\n",_f );
		close( fd );
		return( rc );
	}

	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,optPPPIfname,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t) &ihk;
	if (ioctl( fd,SIOCSHOOKREGISTER,(caddr_t)&ifr ) != 0) {
		rc = errno;
		ERR( "%s ioctl(SIOCSHOOKREGISTER) failed\n",_f );
		close( fd );
		return( rc );
	}
	LOG( "kernel mode PPPoE enabled for %s\n",optPPPIfname);
	close( fd );
	return( 0 );
}

static void
kernel_session_finit( PPPoEConnection *conn )
{
	static char	*_f = "kernel_session_finit:";
	int		fd;
	struct ifreq	ifr;
	struct ifhook	ihk;

	if ((fd = socket( PF_INET,SOCK_DGRAM,0 )) < 0) {
		ERR( "%s open socket\n",_f );
		return;
	}
	memset ( &ihk,'\0',sizeof( ihk ));
	strncpy( ihk.ihk_name,"pppoe-session-recv",sizeof( ihk.ihk_name ));
	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,optPPPIfname,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t) &ihk;
	if (ioctl( fd,SIOCGHOOK,(caddr_t)&ifr ) != 0) {
		ERR( "%s ioctl(SIOCGHOOK) failed\n",_f );
		close( fd );
		return;
	}
	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,conn->ifName,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t) &ihk;
	if (ioctl( fd,SIOCSHOOKUNREGISTER,(caddr_t)&ifr ) != 0) {
		ERR( "%s ioctl(SIOCSHOOKUNREGISTER) failed\n",_f );
		close( fd );
		return;
	}

	memset ( &ihk,'\0',sizeof( ihk ));
	strncpy( ihk.ihk_name,"pppoe-session-xmit",sizeof( ihk.ihk_name ));
	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,optPPPIfname,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t) &ihk;
	if (ioctl( fd,SIOCGHOOK,(caddr_t)&ifr ) != 0) {
		ERR( "%s ioctl(SIOCGHOOK) failed\n",_f );
		close( fd );
		return;
	}
	memset ( &ifr,'\0',sizeof( ifr ));
	strncpy( ifr.ifr_name,optPPPIfname,sizeof( ifr.ifr_name ));
	ifr.ifr_data = (caddr_t) &ihk;
	if (ioctl( fd,SIOCSHOOKUNREGISTER,(caddr_t)&ifr ) != 0) {
		ERR( "%s ioctl(SIOCSHOOKUNREGISTER) failed\n",_f );
		close( fd );
		return;
	}
	close( fd );
}
#endif


/***********************************************************************
*%FUNCTION: sigPADT
*%ARGUMENTS:
* src -- signal received
*%RETURNS:
* Nothing
*%DESCRIPTION:
* If an established session exists send PADT to terminate from session
*  from our end
***********************************************************************/
void
sigPADT(int src)
{
  LOG( "received signal %d\n",(int)src );
  sendPADT( Connection,"pppoed: received signal" );
  doExit( EXIT_SUCCESS );
}

/**********************************************************************
*%FUNCTION: usage
*%ARGUMENTS:
* argv0 -- program name
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Prints usage information and exits.
***********************************************************************/
void
usage(char const *argv0)
{
    fprintf(stderr, "Usage: %s [options]\n", argv0);
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "   -I if_name     -- Specify interface (REQUIRED)\n");
    fprintf(stderr, "   -T timeout     -- Specify inactivity timeout in seconds\n");
    fprintf(stderr, "   -D filename    -- Log debugging information in filename\n");
    fprintf(stderr, "   -V             -- Print version and exit\n");
    fprintf(stderr, "   -A             -- Print access concentrator names and exit\n");
    fprintf(stderr, "   -S name        -- Set desired service name\n");
    fprintf(stderr, "   -C name        -- Set desired access concentrator name\n");
    fprintf(stderr, "   -U             -- Use Host-Unique to allow multiple PPPoE sessions\n");
    fprintf(stderr, "   -s             -- Use synchronous PPP encapsulation\n");
    fprintf(stderr, "   -m MSS         -- Clamp incoming and outgoing MSS options\n");
#ifdef WITH_PIDFILE
    fprintf(stderr, "   -p pidfile     -- Write process-ID to pidfile\n");
#endif
    fprintf(stderr, "   -e sess:mac    -- Skip discovery phase; use existing session\n");
    fprintf(stderr, "   -n             -- Do not open discovery socket\n");
    fprintf(stderr, "   -k             -- Kill a session with PADT (requires -e)\n");
    fprintf(stderr, "   -d             -- Perform discovery, print session info and exit\n");
    fprintf(stderr, "   -f disc:sess   -- Set Ethernet frame types (hex)\n");
#ifdef WITH_KERNEL_MODE_PPPOE
    fprintf(stderr, "   -x             -- Enable kernel mode pppoe (session only)\n");
#endif
    fprintf(stderr, "   -h             -- Print usage information\n");
    doExit(EXIT_SUCCESS);
}

/**********************************************************************
*%FUNCTION: main
*%ARGUMENTS:
* argc, argv -- count and values of command-line arguments
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Main program
***********************************************************************/
int
main(int argc, char *argv[], char *env[])
{
    int opt;
    int n;
    unsigned int m[6];		/* MAC address in -e option */
    unsigned int s;		/* Temporary to hold session */
#ifdef WITH_PIDFILE
    FILE *pidfile;
#endif
    unsigned int discoveryType, sessionType;

    PPPoEConnection conn;

#ifdef HAVE_N_HDLC
    int disc = N_HDLC;
    long flags;
#endif

    /* Initialize syslog */
    OPENLOG( "pppoed",LOG_PID|LOG_CONS,LOG_DAEMON );
    LOG( "version %s started\n",VERSION" (" __DATE__ ", " __TIME__ ")" );

#if 0
    if (nice( -19 ) < 0) {	/* XXX jse */
	ERR( "nice failed\n" );
	doExit( 1 );
    }
#endif

    /* Initialize connection info */
    memset(&conn, 0, sizeof(conn));
    conn.discoverySocket = -1;
    conn.sessionSocket = -1;

    /* For signal handler */
    Connection = &conn;

    while((opt = getopt(argc, argv, "I:VAT:D:hS:C:Usm:np:e:kdf:F:x")) != -1) {
	switch(opt) {
	case 'F':
#ifdef WITH_FLOOD_DISCOVERY
	    if (sscanf(optarg, "%d", &optFloodDiscovery) != 1) {
		fprintf(stderr, "Illegal argument to -F: Should be -F numFloods\n");
		doExit(EXIT_FAILURE);
	    }
	    if (optFloodDiscovery < 1) optFloodDiscovery = 1;
	    fprintf(stderr, "WARNING: DISCOVERY FLOOD IS MEANT FOR STRESS-TESTING\n");
	    fprintf(stderr, "A PPPOE SERVER WHICH YOU OWN.  DO NOT USE IT AGAINST\n");
	    fprintf(stderr, "A REAL ISP.  YOU HAVE 5 SECONDS TO ABORT.\n");
	    sleep(5);
#else
	    optFloodDiscovery = 0;
#endif
	    break;
	case 'f':
	    if (sscanf(optarg, "%x:%x", &discoveryType, &sessionType) != 2) {
		ERR( "illegal argument to -f: Should be disc:sess in hex\n" );
		doExit(EXIT_FAILURE);
	    }
	    Eth_PPPOE_Discovery = (UINT16_t) discoveryType;
	    Eth_PPPOE_Session   = (UINT16_t) sessionType;
	    break;
	case 'd':
	    optSkipSession = 1;
	    break;

	case 'k':
	    conn.killSession = 1;
	    break;

	case 'n':
	    /* Do not even open a discovery socket -- used when invoked 
	       by pppoe-server */
	    conn.noDiscoverySocket = 1;
	    break;

	case 'e':
	    /* Existing session: "sess:xx:yy:zz:aa:bb:cc" where "sess" is
	       session-ID, and xx:yy:zz:aa:bb:cc is MAC-address of peer */
	    n = sscanf(optarg, "%u:%2x:%2x:%2x:%2x:%2x:%2x",
		       &s, &m[0], &m[1], &m[2], &m[3], &m[4], &m[5]);
	    if (n != 7) {
		ERR( "illegal argument to -e: Should be sess:xx:yy:zz:aa:bb:cc\n" );
		doExit(EXIT_FAILURE);
	    }

	    /* Copy MAC address of peer */
	    for (n=0; n<6; n++) {
		conn.peerEth[n] = (unsigned char) m[n];
	    }

	    /* Convert session */
	    conn.session = htons(s);

	    /* Skip discovery phase! */
	    conn.skipDiscovery = 1;
	    break;

	case 'p':
#ifdef WITH_PIDFILE
	    pidfile = fopen(optarg, "w");
	    if (pidfile) {
		fprintf(pidfile, "%lu\n", (unsigned long) getpid());
		fclose(pidfile);
	    }
#endif
	    break;
	case 'S':
	    SET_STRING(conn.serviceName, optarg);
	    break;
	case 'C':
	    SET_STRING(conn.acName, optarg);
	    break;
	case 's':
	    conn.synchronous = 1;
	    break;
	case 'U':
	    conn.useHostUniq = 1;
	    break;
	case 'D':
#ifdef DUMP_FRAMES
	    conn.debugFile = fopen(optarg, "w");
	    if (!conn.debugFile) {
		fprintf(stderr, "Could not open %s: %s\n",
			optarg, strerror(errno));
		doExit(EXIT_FAILURE);
	    }
	    fprintf(conn.debugFile, "rp-pppoe-%s\n", VERSION);
	    fflush(conn.debugFile);
#else
	    conn.debugFile = NULL;
#endif
	    break;
	case 'T':
	    optInactivityTimeout = (int) strtol(optarg, NULL, 10);
	    if (optInactivityTimeout < 0) {
		optInactivityTimeout = 0;
	    }
	    break;
	case 'm':
	    optClampMSS = (int) strtol(optarg, NULL, 10);
	    if (optClampMSS < 536) {
		ERR( "-m: %d is too low (min 536)\n",optClampMSS );
		doExit(EXIT_FAILURE);
	    }
	    if (optClampMSS > 1452) {
		ERR( "-m: %d is too high (max 1452)\n",optClampMSS );
		doExit(EXIT_FAILURE);
	    }
	    break;
	case 'I':
	    SET_STRING(conn.ifName, optarg);
	    break;
	case 'V':
	    LOG( "Roaring Penguin PPPoE Version %s\n",VERSION );
	    doExit(EXIT_SUCCESS);
	case 'A':
	    conn.printACNames = 1;
	    break;
#ifdef WITH_KERNEL_MODE_PPPOE
	case 'x':
		if ((optPPPIfname = getenv( "IFNAME" )) != NULL) {
			optKernelModePPPoE = 1;
		} else {
			ERR( "-x: no environment IFNAME found\n" );
			doExit(EXIT_FAILURE);
		}
		break;
#endif
	case 'h':
	default:
	    usage(argv[0]);
	    break;
	}
    }

    if (!conn.ifName) {
	ERR( "no interface specified (-I option)\n" );
	doExit(EXIT_FAILURE);
    }

    /* Set signal handlers: send PADT on TERM, HUP and INT */
    if (!conn.printACNames) {
	signal(SIGTERM, sigPADT);
	signal(SIGHUP, sigPADT);
	signal(SIGINT, sigPADT);

#ifdef HAVE_N_HDLC
	if (conn.synchronous) {
	    if (ioctl(0, TIOCSETD, &disc) < 0) {
		ERR( "unable to set line discipline to N_HDLC -- synchronous mode probably will fail\n" );
	    } else {
		INF( "changed pty line discipline to N_HDLC for synchronous mode\n" );
	    }
	    /* There is a bug in Linux's select which returns a descriptor
	     * as readable if N_HDLC line discipline is on, even if
	     * it isn't really readable.  This return happens only when
	     * select() times out.  To avoid blocking forever in read(),
	     * make descriptor 0 non-blocking */
	    flags = fcntl(0, F_GETFL);
	    if (flags < 0) fatalSys("fcntl (F_GETFL)");
	    if (fcntl(0, F_SETFL, (long) flags | O_NONBLOCK) < 0) {
		fatalSys("fcntl (F_SETFL)");
	    }
	}
#endif

    }

#ifdef WITH_FLOOD_DISCOVERY
    if (optFloodDiscovery) {
	for (n=0; n < optFloodDiscovery; n++) {
	    if (conn.printACNames) {
		LOG( "sending discovery flood %d\n",n+1 );
	    }
	    discovery(&conn);
	    conn.discoveryState = STATE_SENT_PADI;
	    close(conn.discoverySocket);
	}
	doExit(EXIT_SUCCESS);
    }
#endif

    discovery(&conn);
    if (optSkipSession) {
	printf("%u:%02x:%02x:%02x:%02x:%02x:%02x\n",
	       ntohs(conn.session),
	       conn.peerEth[0],
	       conn.peerEth[1],
	       conn.peerEth[2],
	       conn.peerEth[3],
	       conn.peerEth[4],
	       conn.peerEth[5]);
	doExit(EXIT_SUCCESS);
    }
#ifdef WITH_KERNEL_MODE_PPPOE
	if (optKernelModePPPoE) {
		if (kernel_session( &conn,optClampMSS ) == 0) {
			optPPPoEConn = &conn;
		} else {
			optKernelModePPPoE = 0;
		}
	}
#endif
    session(&conn);
    doExit(EXIT_SUCCESS);
    return 0;
}

/**********************************************************************
*%FUNCTION: fatalSys
*%ARGUMENTS:
* str -- error message
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Prints a message plus the errno value to stderr and syslog and exits.
***********************************************************************/
void
fatalSys(char const *str)
{
    char buf[1024];
    sprintf(buf, "%.256s errno=%.256s", str, strerror(errno));
    ERR( "%s\n",buf );
    sendPADT( Connection,"pppoed: system call error" );
    doExit( errno );
}

/**********************************************************************
*%FUNCTION: sysErr
*%ARGUMENTS:
* str -- error message
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Prints a message plus the errno value to syslog.
***********************************************************************/
void
sysErr(char const *str)
{
    char buf[1024];
    sprintf(buf, "%.256s errno=%.256s", str, strerror(errno));
    ERR( "%s\n",buf );
}

/**********************************************************************
*%FUNCTION: rp_fatal
*%ARGUMENTS:
* str -- error message
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Prints a message to stderr and syslog and exits.
***********************************************************************/
void
rp_fatal(char const *str)
{
    char buf[1024];
    ERR( "%s\n",str );
    sprintf(buf, "pppoed: %.256s", str);
    sendPADT(Connection, buf);
    doExit( EXIT_FAILURE );
}

/**********************************************************************
*%FUNCTION: asyncReadFromEth
*%ARGUMENTS:
* conn -- PPPoE connection info
* sock -- Ethernet socket
* clampMss -- if non-zero, do MSS-clamping
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Reads a packet from the Ethernet interface and sends it to async PPP
* device.
***********************************************************************/
void
asyncReadFromEth(PPPoEConnection *conn, int sock, int clampMss)
{
    PPPoEPacket packet;
    int len;
    int plen;
    int i;
    unsigned char pppBuf[4096];
    unsigned char *ptr = pppBuf;
    unsigned char c;
#ifndef FAKE_PPP_FCS
    UINT16_t fcs;
    unsigned char header[2] = {FRAME_ADDR, FRAME_CTRL};
    unsigned char tail[2];
#endif
#ifdef USE_BPF
    int type;
#endif

    if (receivePacket(sock, &packet, &len) < 0) {
	return;
    }

    /* Check length */
    if (len < HDR_SIZE) {
	ERR( "bogus PPPoE packet length (%d)\n",len );
	return;
    }
    if (ntohs(packet.length) + HDR_SIZE > len) {
	ERR( "bogus PPPoE length field (%u)\n",
	       (unsigned int) ntohs(packet.length));
	return;
    }
#ifdef DUMP_FRAMES
    if (conn->debugFile) {
	dumpPacket(conn->debugFile, &packet, "RCVD");
	fprintf(conn->debugFile, "\n");
	fflush(conn->debugFile);
    }
#endif

#ifdef USE_BPF
    /* Make sure this is a session packet before processing further */
    type = etherType(&packet);
    if (type == Eth_PPPOE_Discovery) {
	sessionDiscoveryPacket(&packet);
    } else if (type != Eth_PPPOE_Session) {
	return;
    }
#endif

    /* Sanity check */
    if (packet.code != CODE_SESS) {
	ERR( "unexpected packet code %d\n",(int)packet.code );
	return;
    }
    if (packet.ver != 1) {
	ERR( "unexpected packet version %d\n",(int)packet.ver );
	return;
    }
    if (packet.type != 1) {
	ERR( "unexpected packet type %d\n",(int)packet.type );
	return;
    }
    if (memcmp(packet.ethHdr.h_dest, conn->myEth, ETH_ALEN)) {
	return;
    }
    if (memcmp(packet.ethHdr.h_source, conn->peerEth, ETH_ALEN)) {
	/* Not for us -- must be another session.  This is not an error,
	   so don't log anything.  */
	return;
    }

    if (packet.session != conn->session) {
	/* Not for us -- must be another session.  This is not an error,
	   so don't log anything.  */
	return;
    }
    plen = ntohs(packet.length);
    if (plen + HDR_SIZE > len) {
	ERR( "bogus length field in session packet %d (%d)\n",
	       (int) plen, (int) len);
	return;
    }

    /* Clamp MSS */
    if (clampMss) {
	clampMSS(&packet, "incoming", clampMss);
    }

#ifndef FAKE_PPP_FCS
    /* Compute FCS */
    fcs = pppFCS16(PPPINITFCS16, header, 2);
    fcs = pppFCS16(fcs, packet.payload, plen) ^ 0xffff;
    tail[0] = fcs & 0x00ff;
    tail[1] = (fcs >> 8) & 0x00ff;
#endif

    /* Build a buffer to send to PPP */
    *ptr++ = FRAME_FLAG;
    *ptr++ = FRAME_ADDR;
    *ptr++ = FRAME_ESC;
    *ptr++ = FRAME_CTRL ^ FRAME_ENC;

    for (i=0; i<plen; i++) {
	c = packet.payload[i];
	if (c == FRAME_FLAG || c == FRAME_ADDR || c == FRAME_ESC || c < 0x20) {
	    *ptr++ = FRAME_ESC;
	    *ptr++ = c ^ FRAME_ENC;
	} else {
	    *ptr++ = c;
	}
    }
#ifndef FAKE_PPP_FCS
    for (i=0; i<2; i++) {
	c = tail[i];
	if (c == FRAME_FLAG || c == FRAME_ADDR || c == FRAME_ESC || c < 0x20) {
	    *ptr++ = FRAME_ESC;
	    *ptr++ = c ^ FRAME_ENC;
	} else {
	    *ptr++ = c;
	}
    }
#else
    *ptr++ = 0xaf;
    *ptr++ = 0xfe;
#endif
    *ptr++ = FRAME_FLAG;

    /* Ship it out */
    if (write(1, pppBuf, (ptr-pppBuf)) < 0) {
	fatalSys("asyncReadFromEth: write failed");
    }
}

/**********************************************************************
*%FUNCTION: syncReadFromEth
*%ARGUMENTS:
* conn -- PPPoE connection info
* sock -- Ethernet socket
* clampMss -- if true, clamp MSS.
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Reads a packet from the Ethernet interface and sends it to sync PPP
* device.
***********************************************************************/
void
syncReadFromEth(PPPoEConnection *conn, int sock, int clampMss)
{
    PPPoEPacket packet;
    int len;
    int plen;
    struct iovec vec[2];
    unsigned char dummy[2];
#ifdef USE_BPF
    int type;
#endif

    if (receivePacket(sock, &packet, &len) < 0) {
	return;
    }

    /* Check length */
    if (ntohs(packet.length) + HDR_SIZE > len) {
	ERR( "bogus PPPoE length field (%u)\n",
	       (unsigned int) ntohs(packet.length));
	return;
    }
#ifdef DUMP_FRAMES
    if (conn->debugFile) {
	dumpPacket(conn->debugFile, &packet, "RCVD");
	fprintf(conn->debugFile, "\n");
	fflush(conn->debugFile);
    }
#endif

#ifdef USE_BPF
    /* Make sure this is a session packet before processing further */
    type = etherType(&packet);
    if (type == Eth_PPPOE_Discovery) {
	sessionDiscoveryPacket(&packet);
    } else if (type != Eth_PPPOE_Session) {
	return;
    }
#endif

    /* Sanity check */
    if (packet.code != CODE_SESS) {
	ERR( "unexpected packet code %d\n",(int)packet.code );
	return;
    }
    if (packet.ver != 1) {
	ERR( "unexpected packet version %d\n",(int)packet.ver );
	return;
    }
    if (packet.type != 1) {
	ERR( "unexpected packet type %d\n",(int)packet.type );
	return;
    }
    if (memcmp(packet.ethHdr.h_dest, conn->myEth, ETH_ALEN)) {
	/* Not for us -- must be another session.  This is not an error,
	   so don't log anything.  */
	return;
    }
    if (memcmp(packet.ethHdr.h_source, conn->peerEth, ETH_ALEN)) {
	/* Not for us -- must be another session.  This is not an error,
	   so don't log anything.  */
	return;
    }
    if (packet.session != conn->session) {
	/* Not for us -- must be another session.  This is not an error,
	   so don't log anything.  */
	return;
    }
    plen = ntohs(packet.length);
    if (plen + HDR_SIZE > len) {
	ERR( "bogus length field in session packet %d (%d)\n",
	       (int) plen, (int) len);
	return;
    }

    /* Clamp MSS */
    if (clampMss) {
	clampMSS(&packet, "incoming", clampMss);
    }

    /* Ship it out */
    vec[0].iov_base = (void *) dummy;
    dummy[0] = FRAME_ADDR;
    dummy[1] = FRAME_CTRL;
    vec[0].iov_len = 2;
    vec[1].iov_base = (void *) packet.payload;
    vec[1].iov_len = plen;

    if (writev(1, vec, 2) < 0) {
	fatalSys("syncReadFromEth: write failed");
    }
}
