/*
 *
 */

#define BUZZER_VERSION		"0.01"



#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <asm/hardware.h>
#include "buzzer.h"

/*
 *	We sponge a minor off of the misc major. No need slurping
 *	up another valuable major dev number for this.
 */

#define BUZZER_MINOR	182
/*
	 BUZZER_WAVE - PB1 timer1 wawe output
*/


#define DEFAULT_HZ 2093
	
#define EVENT_TICKS ( HZ / 100 )    // roughly 10 ms per note

#define BUZZSIZ 4096

/*
 * this device receives a string and plays chars:
 *
 * ' ' = silence
 * '+' = repeat string
 * 'a' = 500 Hz
 * 'b' = 1280 Hz
 * 'c' = 2060 Hz
 * 'd' = 2840 Hz
 * 'e' = 3620 Hz
 * 'f' = 4400 Hz
 * 'g' = 5180 Hz
 * 'h' = 5960 Hz
 * 'i' = 6740 Hz
 * 'j' = 7520 Hz
 * 'k' = 8300 Hz
 * 'l' = 9080 Hz
 * 'm' = 9860 Hz
 * 'n' = 10640 Hz
 * 'o' = 11420 Hz
 * 'p' = 12200 Hz
 * 'q' = 12980 Hz
 * 'r' = 13760 Hz
 * 's' = 14540 Hz
 * 't' = 15320 Hz
 * 'u' = 16100 Hz
 * 'v' = 16880 Hz
 * 'w' = 17660 Hz
 * 'x' = 18440 Hz
 * 'y' = 19220 Hz
 * 'z' = 20000 Hz
 *
 * you can prefix letters with digits to multiply duration:
 *
 * "100 a2b" is a 1s silence, 10ms 500Hz tone and 20ms 1280Hz tone
 *
 */

   
volatile struct trio_timers* tt; 
volatile struct trio_timer_channel* tc;

static int buzzer_ticks;             // remaining ticks before event
static int buzzer_stop;              // tell the interrupt to stop playing
static struct wait_queue *buzzer_wq; // writers wait queue
static struct tq_struct buzzer_tq;   // event queue
static int buzzer_nbnotes;           // remaining notes to play
static char* buzzer_notes;           // pointer to next notes
static char* buzzer_buffer;          // notes buffer

static void buzzer_play(void * data)
{
	char note;
	static char lastnote;
	if(buzzer_stop || (buzzer_nbnotes <= 0)) // we have no more notes to play, or we're told to stop
	{
		kfree(buzzer_buffer); // free the buffer
		buzzer_buffer = 0; // and mark it as freed
		tc->ccr = TC_CLKDIS; // stop buzzer
		buzzer_nbnotes = 0; // tell writers we're done playing
		buzzer_ticks = 0; // reset duration counter
		wake_up_interruptible(&buzzer_wq); // wake_up next writer
		return;
	}
	queue_task(&buzzer_tq, &tq_timer); // requeue me
	if(buzzer_ticks--) // we effectively play only each EVENT_TICKS ticks
		return;
	buzzer_ticks = EVENT_TICKS;
	note = *(buzzer_notes++);  // fetch next note
	buzzer_nbnotes--;
	if(note == lastnote)
		return;
	if(note >= '0' && note <= '9') // a multiplicator value
	{
		int cmult = 0;
		while(buzzer_notes && (note >= '0' && note <= '9'))
		{
			cmult = 10 * cmult + (note - '0');
			note = *(buzzer_notes++);
			buzzer_nbnotes--;
		}
		buzzer_ticks *= cmult;
	}
	if(note == ' ') // silence
		tc->ccr = TC_CLKDIS; // stop buzzer
	else if(note == '+') // repeat all
	{
		buzzer_nbnotes += (buzzer_notes - buzzer_buffer);
		buzzer_notes = buzzer_buffer;
		return;
	}
	else if(note >= 'a' && note <= 'z') // a note
	{
		int freq = ((note - 'a') * 780 + 500);
		tc->rc = (ARM_CLK / (2*freq)) - 1;  // set frequency
		tc->ccr = TC_SWTRG |TC_CLKEN; // start buzzer
	}
	lastnote = note;
}

static int buzzer_write(struct inode *inode, struct file *file, const char *buf,
					int count)
{
	buzzer_stop = 1;
	if(file->f_flags & O_NONBLOCK)
	{
		if(buzzer_nbnotes)
			return -EAGAIN;
	}
	else // avoid the racy while(lock) sleep_on();
	{
		struct wait_queue wait = { current, NULL };
		add_wait_queue(&buzzer_wq, &wait);
		current->state = TASK_INTERRUPTIBLE;
		while(buzzer_nbnotes)
		{
			schedule();
			if(current->signal & ~current->blocked)
				return -ERESTARTSYS;
		}
		remove_wait_queue(&buzzer_wq, &wait);
	}
	if(count && !(buzzer_buffer = kmalloc(count, GFP_KERNEL)))
		return -ENOMEM;
	buzzer_nbnotes = count;
	buzzer_notes = buzzer_buffer;
	memcpy_fromfs(buzzer_buffer, buf, buzzer_nbnotes);
	buzzer_stop = 0;
	queue_task(&buzzer_tq, &tq_timer);
	return buzzer_nbnotes;
}

static int buzzer_open(struct inode *inode, struct file *file)
{
	return 0;
}
static void buzzer_release(struct inode *inode, struct file *file)
{
}

static struct file_operations buzzer_fops = {
	NULL,
	NULL,
	buzzer_write,
	NULL,		/* No readdir */
	NULL,
	NULL,
	NULL,		/* No mmap */
	buzzer_open,
	buzzer_release
};

static struct miscdevice buzzer_dev=
{
	BUZZER_MINOR,
	"buzzer",
	&buzzer_fops
};


int buzzer_init(void)
{
	unsigned long v;

	buzzer_tq.routine = buzzer_play;

	tt = (struct trio_timers*) (PERIPH_BASE+TC_OFFSET);
	tc = &tt->chans[1].ch;
	
	/* program NO signal on XC1 */
	v = tt->bmr;
	v = (v & ~(3 << 2)) | (1 << 2);
	tt->bmr = v;

	tc->ccr = TC_CLKDIS;  /* disable the channel */

	/* select ACLK/2 as input frequency for TC1 and enable Wave */
	/* eevt = TIOB, ENETRG = 0, CPTRG = 1, WAVE=1, ACPA= none, ACPC = toggle*/
	/* BCPC = none, BEEVT=none, BSWTRG=none CPCDIS=0*/
	tc->cmr = (1 << 15) |(1 << 14)| (3 << 18) ;

	tc->idr = ~0ul;  /* disable all interrupt */
	tc->rc = (ARM_CLK/(2*DEFAULT_HZ)) - 1;

	tc->ier = 0;  /* enable CPCS interrupt */
	misc_register(&buzzer_dev);
	return 0;
}
