/*======================================================================

  $Id: mtd.c,v 1.12 2008-05-19 06:44:53 jselent Exp $

    A general driver for accessing PCMCIA card memory via Bulk
    Memory Services.

    This driver provides the equivalent of /dev/mem for a PCMCIA
    card's attribute and common memory.  It includes character
    and block devices.

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/module.h>

#include <linux/sysdep-2.1.h>

#if 0
#include <asm/uaccess.h>
#endif
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <stdarg.h>

#include <linux/mtd.h>

/* Major device #'s for memory device */

static struct mtd_info *mtd_table[MAX_MTD_DEVICES];



static int mtd_lseek(struct inode *inode, struct file *file, off_t offset, 
					   int origin);

static int mtd_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg);
static int mtd_read(struct inode *inode, struct file *file, char *buf, int count);
static int mtd_write(struct inode *inode, struct file *file, const char *buf, int count);
static int mtd_open(struct inode *inode, struct file *file);
static void mtd_close(struct inode *inode,    struct file *file);
static int mtd_fsync(struct inode *inode,    struct file *file);

static struct file_operations mtd_fops = {
	mtd_lseek,     	/* lseek */
	mtd_read,	/* read */
	mtd_write, 	/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
	mtd_ioctl,	/* ioctl */
	NULL,		/* mmap */
	mtd_open,	/* open */
	mtd_close,	/* release */
	mtd_fsync,	/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL		/* revalidate */
};

static struct mtd_notifier *mtd_notifiers = NULL;

static int mtd_lseek(struct inode *inode, struct file *file, off_t offset,  int orig)
{
	int devnum = MINOR(inode->i_rdev) >> 1;
	struct mtd_info *mtd=mtd_table[devnum];

	if (devnum >=MAX_MTD_DEVICES || !mtd) 
		return -ENODEV;

	switch (orig) {
	case 0: 
		/* SEEK_SET */
		file->f_pos = offset;
		break;
	case 1: 
		/* SEEK_CUR */
		file->f_pos += offset;
		break;
	case 2:
		/* SEEK_END */
		file->f_pos =mtd->size + offset;
		break;
	default: 
		return -EINVAL;
	}

	if (file->f_pos < 0) 
		file->f_pos = 0;
	else if (file->f_pos >= mtd->size)
		file->f_pos = mtd->size - 1;

	return file->f_pos;
}

static int mtd_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	int devnum = minor >> 1;
	struct mtd_info *mtd=mtd_table[devnum];	

	if (devnum >= MAX_MTD_DEVICES || !mtd)
		return -ENODEV;
	
	if (file->f_mode & 2)
	{
		/* We're trying to open it with write access. */
		/* Fail if it's read-only, or if we're using the ro device */
		if ((minor & 1) || !(mtd->flags & MTD_WRITEABLE))
			return -EACCES;
	}

	/* OK. We're going to open it now. */

	mtd->usage++;	
	(*(mtd->openclose))(mtd, 1);

	return 0;
} /* mtd_open */

/*====================================================================*/

static void mtd_close( struct inode *inode,struct file *file )
{
	int minor = MINOR(inode->i_rdev);
	struct mtd_info *mtd;
	int devnum = minor >> 1;

	if (devnum >= MAX_MTD_DEVICES)
		return;

	mtd = mtd_table[devnum];
	
	if (!mtd) return;
	MTD_SYNC(mtd);
	(*(mtd->openclose))(mtd,0);
	mtd->usage--;	  
} /* mtd_close */

static int mtd_fsync(struct inode *inode,  struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct mtd_info *mtd;
	int devnum = minor >> 1;

	if (devnum >= MAX_MTD_DEVICES)
		return -ENODEV;

	mtd = mtd_table[devnum];

	if (!mtd) return -ENODEV;
	MTD_SYNC(mtd);

	return 0;
} /* mtd_close */



static int mtd_read(struct inode* inode, struct file *file, char *buf, int count)
{
	int devnum = MINOR(inode->i_rdev) >> 1;
	struct mtd_info *mtd = mtd_table[devnum];
	u_char *mtdbuf;
	u_long retlen=0;
	u_long count1;
	int ret=0;
	u_long pos = file->f_pos;
	
	if (!mtd) return -ENODEV;

	if (pos + count > mtd->size)
		count = mtd->size - pos;

	if (!count)
		printk(KERN_NOTICE "Count is zero.\n");

	count1 = count;

	while (count && !ret)
	{
		ret = MTD_POINT(mtd, pos, count, &retlen, &mtdbuf);
#warning ERROR XXX FSchaller in mtd.c ???XXX
#if 0
		if (ret)
		{
			/* mtd->point() failed; use mtd->read instead */
#endif		
		ret = MTD_READ(mtd, pos, count, &retlen, buf, 1);
		if (retlen)
		{
			file->f_pos += retlen;
			return retlen;
		}
		else 
			return ret;
#if 0
		}
#endif
		if (!retlen)
		{
			printk(KERN_NOTICE "mtd_read: device returned no error, but no data either!\n");
			ret = -EIO;
			break;
		}
		
		if (copy_to_user(buf,mtdbuf,retlen) < 0) 
			ret = -EFAULT;
		else
		{
			count -= retlen;
			file->f_pos += retlen;
		}
		MTD_UNPOINT(mtd, pos);
	}
	
	if (count < count1)
		return count1-count;
	else 
		return ret;
} /* mtd_read */


static int mtd_write(struct inode* inode, struct file *file, const char *buf, int count)
{
	int devnum = MINOR(inode->i_rdev) >> 1;
	struct mtd_info *mtd = mtd_table[devnum];
	u_long count1;
	u_long retlen=0;
	int ret=0;
	u_long pos = file->f_pos;
	u_char *mtdbuf;
	
	if (!mtd) return -ENODEV;

	if (pos == mtd->size)
		return -ENOSPC;
	
	if (pos + count > mtd->size)
		count = mtd->size - pos;
	
	count1 = count;
	
	while (count)
	{
#if 1
		if (file->f_flags & O_NONBLOCK)
		{
			ret = MTD_POINT(mtd, pos, count, &retlen, &mtdbuf);
			MTD_UNPOINT(mtd, pos);
			if (ret) {
				return -EAGAIN;
			}
		}
#endif
		ret = (*(mtd->write))(mtd, pos, count, &retlen, buf, 1);

		if (ret)
			break;

		if (!retlen)
		{
			printk(KERN_NOTICE "mtd_write: device returned no error, but no data written\n");
			ret = -EIO;
			break;
		}

		count -= retlen;
	}

	if (count < count1)
	{
		file->f_pos += count1-count;
		return count1-count;
	}
	else 
		return ret;
} /* mtd_write */

/*======================================================================

    IOCTL calls for getting device parameters.

======================================================================*/

int
reload_mtd_device( struct mtd_info *mtd,int devnum )
{
	struct mtd_notifier *not=mtd_notifiers;

	while (not) {
		(*(not->func))(NULL,devnum);
		(*(not->func))(mtd,devnum);
		not = not->next;
	}
	return (0);
}

static int mtd_ioctl(struct inode *inode, struct file *file,
		       u_int cmd, u_long arg)
{
	int devnum = MINOR(inode->i_rdev) >> 1;
	struct mtd_info *mtd=mtd_table[devnum];
	int ret = 0;
	u_long size;
	

	if (devnum >= MAX_MTD_DEVICES || !mtd)
		return -ENODEV;
	
	size = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
	if (cmd & IOC_IN) {
		ret = verify_area(VERIFY_READ, (char *)arg, size);
		if (ret) return ret;
	}
	if (cmd & IOC_OUT) {
		ret = verify_area(VERIFY_WRITE, (char *)arg, size);
		if (ret) return ret;
	}
	
	switch (cmd) {
	case MEMGETINFO:
		copy_to_user((struct mtd_info *)arg, mtd,
			     (u_long) &mtd->erase - (u_long)mtd);
		break;
	case MEMERASE:
	{
		struct erase_info *erase=kmalloc(sizeof(struct erase_info),GFP_KERNEL);
		if (!erase)
			ret = -ENOMEM;
		else
		{
			memset (erase,0,sizeof(struct erase_info));
			copy_from_user(&erase->addr, (u_long *)arg,
				       2 * sizeof(u_long));
			erase->mtd = mtd;
			ret = MTD_ERASE(mtd, erase);
		}
		break;
	 
	}

	case MEMADD:
		ret = reload_mtd_device(mtd, devnum);
		break;
	
	default:
		ret = -EINVAL;
	}
	
	return ret;
} /* memory_ioctl */


/*====================================================================*/

int add_mtd_device(struct mtd_info *mtd)
{
	int i;

	for (i=0; i< MAX_MTD_DEVICES; i++)
		if (!mtd_table[i])
		{
			struct mtd_notifier *not=mtd_notifiers;

			mtd_table[i]=mtd;
			
			printk(KERN_NOTICE "mtd: Giving out device %d to %s\n",i, mtd->name);
			while (not)
			{
				(*(not->func))(mtd,i);
				not = not->next;
			}
			MOD_INC_USE_COUNT;
			return 0;
		}
	
	return 1;
}

struct mtd_info *get_mtd_device(u_char num)
{
	return mtd_table[num];
}

int del_mtd_device (struct mtd_info *mtd)
{
	struct mtd_notifier *not=mtd_notifiers;
	int i;
	
	if (mtd->usage)
	    return -EAGAIN;

	for (i=0; i < MAX_MTD_DEVICES; i++)
	{
		if (mtd_table[i] == mtd)
		{
			while (not)
			{
				(*(not->func))(NULL,i);
				not=not->next;
			}
			mtd_table[i] = NULL;
			MOD_DEC_USE_COUNT;

			return 0;
		}
	}

	return 1;
}



void register_mtd_notifier (struct mtd_notifier *new)
{
	new->next = mtd_notifiers;
	mtd_notifiers = new;
	MOD_INC_USE_COUNT;
}



int unregister_mtd_notifier (struct mtd_notifier *old)
{
	struct mtd_notifier **prev = &mtd_notifiers;
	struct mtd_notifier *cur;

	while ((cur = *prev))
	{
		if (cur == old)
		{
			*prev = cur->next;
			MOD_DEC_USE_COUNT;
			return 0;
		}
		prev = &cur->next;
	}
	return 1;
}


int
mtd_init( void )
{
	int	i;

	for (i=0; i<MAX_MTD_DEVICES; i++) {
		mtd_table[i] = NULL;
	}

	if (register_chrdev( MTD_MAJOR,"mtd",&mtd_fops )) {
		printk( KERN_ERR"Can't allocate major number %d for Memory Technology Devices.\n",
			   MTD_MAJOR);
		return( -EAGAIN );
	}
	return( 0 );
}


/*====================================================================*/

#ifdef MODULE
int init_module(void)
{
	return init_mtd();
}

void cleanup_module(void)
{
	unregister_chrdev(MTD_MAJOR, "mtd");
}
#endif
