/*
 *
 */

#define TRIOPORT_VERSION		"0.01"

/*
 */
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/arch-trio/irq.h>
#include <asm/arch-trio/trioport.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/delay.h>
#include "triopio.h"
/*
 *	We sponge a minor off of the misc major. No need slurping
 *	up another valuable major dev number for this.
 */

#define TRIOPORTA_MINOR	175
#define TRIOPORTB_MINOR	176

static struct trioport porta, portb;
static struct trioport* ports[2];

extern int setup_arm_irq(int, struct irqaction *);

void trioport_io(struct trioport* p, u_32 mask, u_32 value);
static void trioport_irq_on(struct trioport* p, u_32 value);
static void trioport_irq_off(struct trioport* p, u_32 value);
static void trioports_irqs_off(void);

static struct pio_handler *pio_head;
static int handler_id;
static int trioports_initialized=0;

int pio_hook_driver(struct pio_handler *handler){
	struct pio_handler *p;
	unsigned long flags;
	if(!trioports_initialized){
		return -1;
	}
	if (!handler->pio_interrupt){
		// asks only for PIO addresses, no interrupts required
		handler->pa = &porta;
		handler->pb = &portb;
		return 0;
	}
	save_flags_cli(flags);
	if (pio_head == NULL){
		pio_head = handler;
		handler->next = NULL;
		handler->prev = NULL;
		handler_id = 1;
		handler->pa = &porta;
		handler->pb = &portb;
		restore_flags(flags);
		if(handler->pa_mask)
			trioport_irq_on(&porta, handler->pa_mask);
		if(handler->pb_mask)
			trioport_irq_on(&portb, handler->pb_mask);

		return handler_id;
	}
	p = pio_head->next;
	while(p->priority > handler->priority){
		if(p->next){
			p = p->next;
		}else{
			break;
		}
	}
	if(p->next){
		handler->prev = p->prev;
		handler->next = p;
		p->prev->next = handler;
	}else{
		p->next = handler;
		handler->prev = p;
		handler->next = NULL;
	}
	if(handler->pa_mask)
		trioport_irq_on(&porta, handler->pa_mask);
	if(handler->pb_mask)
		trioport_irq_on(&portb, handler->pb_mask);
	
	handler->pa = &porta;
	handler->pb = &portb;

	restore_flags(flags);
	return (++handler_id);
}
int pio_remove_driver(struct pio_handler *handler){
	struct pio_handler *p;
	unsigned long flags;
	save_flags_cli(flags);
	p = pio_head;
	while(p->id != handler->id){
		if(p->next)
			p = p->next;
		else
			break;
	}
	if(p->next){
		p->prev->next = p->next;
		p->next->prev = p->prev;
	}else{
		// end of queue
		if(p->id == handler->id){
			p->prev->next = NULL;
		}else{
			// not found
			return -1;
		}
	}
	if(handler->pa_mask)
		trioport_irq_off(&porta, handler->pa_mask);
	if(handler->pb_mask)
		trioport_irq_off(&portb, handler->pb_mask);

	restore_flags(flags);
	return 0;
}

static int trioport_read(struct inode *inode, struct file *file, char *buffer, int arg){
	struct trioport *p;
	int minor;
	u_32 value;
	if (arg < sizeof(u_32))
		return 0;
	minor = MINOR(inode->i_rdev);
	p = (minor == TRIOPORTA_MINOR)?&porta:&portb; 
	value = trioport_get_in(p);
	put_user(value, buffer);
	return (sizeof(value));
}
u_32 trioport_get_in(struct trioport *p){
	u_32 value;
	unsigned long flags;
	save_flags_cli(flags);
	value = (p->base->odsr & p->outmask) | (p->base->pdsr & p->inmask);
	restore_flags(flags);
	return value;
}
static int trioport_write(struct inode *inode, struct file *file, const char *buffer, int arg){
	struct trioport *p;
	int minor;
	u_32 value,mask;
	if (arg < 2*sizeof(u_32))
		return 0;
	minor = MINOR(inode->i_rdev);
	p = (minor == TRIOPORTA_MINOR)?&porta:&portb;

	mask = *((u_32*)buffer);
	value = *((u_32*)buffer+1);
	trioport_set_out(p, mask, value);
	return 2*sizeof(u_32);
}
int trioport_set_out(struct trioport *p, u_32 mask, u_32 value){
	u_32 status;
	unsigned long flags;
	save_flags_cli(flags);
	p->base->sodr  = value & mask & p->outmask;
	p->base->codr  = ~value & mask & p->outmask;
	status = p->base->odsr;
	restore_flags(flags);
	return status;
}
void trioport_out(struct trioport* p, u_32 value){
	unsigned long flags;
	save_flags_cli(flags);
	p->base->oer = p->pout = value;
	p->base->odr = ~value;
	p->outmask = p->pio & p->base->osr;
	p->inmask = p->pio & ~p->base->osr;
	restore_flags(flags);
}
void trioport_change_out(struct trioport* p, u_32 mask, u_32 value){
	unsigned long flags;
	save_flags_cli(flags);
	p->base->oer = p->pout = value & mask;
	p->base->odr = ~value & mask;
	p->outmask = p->pio & p->base->osr;
	p->inmask = p->pio & ~p->base->osr;
	restore_flags(flags);
}
void trioport_io(struct trioport* p, u_32 mask, u_32 value){
	unsigned long flags;
	save_flags_cli(flags);
	p->base->per = value & mask;
	p->base->pdr = ~value & mask;
	p->pio = p->base->psr;
	restore_flags(flags);
}
static void trioport_irq_on(struct trioport* p, u_32 value){
	p->base->ier = value & p->inmask;
	p->irq = p->base->imr;
	(p->irq)?unmask_irq(p->irq_nb):mask_irq(p->irq_nb);
}
static void trioport_irq_off(struct trioport* p, u_32 value){
	p->base->idr =  value & p->inmask;
	p->irq = p->base->imr;
	(p->irq)?unmask_irq(p->irq_nb):mask_irq(p->irq_nb);
}
static void trioports_irqs_off(){
	unsigned long flags;
	save_flags_cli(flags);
	porta.base->idr = 0xFFFFFFFF;
	porta.irq = porta.base->imr;
	mask_irq(porta.irq_nb);
	portb.base->idr = 0xFFFFFFFF;
	portb.irq = portb.base->imr;
	mask_irq(portb.irq_nb);
	restore_flags(flags);
}
static void pio_interrupt(struct trioport *p){
	u_32 changed;
	struct pio_handler *ph;
	if((ph=pio_head)){
//		changed = p->pdin ^ p->base->pdsr;
		changed = p->base->isr & p->inmask;
		if(changed){
			while(ph){
				if(ph->pio_interrupt(p, changed)){
					break;
				}
				ph = ph->next;
			}
		}
	}
	p->pdin = p->base->pdsr;
}
static void pio_interrupta(int irq, void *dev_id, struct pt_regs * regs){
	pio_interrupt(&porta);
}
static void pio_interruptb(int irq, void *dev_id, struct pt_regs * regs){
	pio_interrupt(&portb);
}
static int trioport_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
						  unsigned long arg)
{
	struct trioport *p;
	int minor = MINOR(inode->i_rdev);
	p = (minor == TRIOPORTA_MINOR)?&porta:&portb;
	switch (cmd) {
		
		case TRIOPORT_IO:
			trioport_io(p, arg, arg);
			break;
			
		case TRIOPORT_CTRL:
			trioport_io(p, arg, ~arg);
			break;

		case TRIOPORT_OUT:
			trioport_out(p, arg);
			break;
		
		case TRIOPORT_IRQ_ON:
			trioport_irq_on(p, arg);
			break;
			
		case TRIOPORT_IRQ_OFF:
			trioport_irq_off(p, arg);
			break;
	
		default:
			return -EINVAL;
	}
	return 0;
}

static int trioport_open(struct inode *inode, struct file *file)
{
	return 0;
}
static void trioport_release(struct inode *inode, struct file *file)
{
}
static struct file_operations trioport_fops = {
	NULL,
	trioport_read,
	trioport_write,
	NULL,		   /* No readdir */
	NULL,
	trioport_ioctl,
	NULL,		/* No mmap */
	trioport_open,
	trioport_release
};
static struct miscdevice trioporta_dev=
{
	TRIOPORTA_MINOR,
	"trioporta",
	&trioport_fops
};
static struct miscdevice trioportb_dev=
{
	TRIOPORTB_MINOR,
	"trioportb",
	&trioport_fops
};

static struct irqaction irq_pioa = { pio_interrupta, 0, 0, "pioa", NULL, NULL};
static struct irqaction irq_piob = { pio_interruptb, 0, 0, "piob", NULL, NULL};

static void interrupts_init(void)
{
	setup_arm_irq(IRQ_PIOA, &irq_pioa);
	setup_arm_irq(IRQ_PIOB, &irq_piob);
}
int trioport_init(void)
{
	porta.base = (struct pio_regs *)PIOA_BASE;
	portb.base = (struct pio_regs *)PIOB_BASE;
	porta.irq_nb = AIC_PIOA;
	portb.irq_nb = AIC_PIOB;
	ports[0] = &porta;
	ports[1] = &portb;

	/* set irqs off, the default PIO config should be already set by boot software */
	trioports_irqs_off();
	interrupts_init();
	
	/* read actual values, set by previous procedures */
	
	porta.pio = porta.base->psr;
	portb.pio = portb.base->psr;
	
	porta.pout = porta.base->odsr;
	portb.pout = portb.base->odsr;

	porta.outmask = porta.pout & porta.pio;
	portb.outmask = portb.pout & portb.pio;

	porta.inmask = ~porta.pout & porta.pio;
	portb.inmask = ~portb.pout & portb.pio;

	porta.id = PIOA_ID;
	portb.id = PIOB_ID;
	
	pio_head = NULL;
	handler_id = 0;
	
	misc_register(&trioporta_dev);
	misc_register(&trioportb_dev);
	trioports_initialized = 1;
	return 0;
}

