/*
 *  linux/arch/arm/kernel/time.c
 *
 *  Copyright (C) 1991, 1992, 1995  Linus Torvalds
 *  Modifications for ARM (C) 1994, 1995, 1996,1997 Russell King
 *
 * This file contains the ARM-specific time handling details:
 * reading the RTC at bootup, etc...
 *
 * 1994-07-02  Alan Modra
 *             fixed set_rtc_mmss, fixed time.year for >= 2000, new mktime
 * 1997-09-10  Updated NTP code according to technical memorandum Jan '96
 *             "A Kernel Model for Precision Timekeeping" by Dave Mills
 */

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 * 
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/timex.h>
#include <asm/hardware.h>
#include <asm/arch/time.h>

#include "bsp.h"
#include "conf.h"

#undef BCD_TO_BIN
#define BCD_TO_BIN(val) (((val)&15) + ((val)>>4)*10)

#undef BIN_TO_BCD
#define BIN_TO_BCD(val) ((((val)/10)<<4) + (val)%10)

extern int		setup_arm_irq(int, struct irqaction *);
static unsigned long	mktime( unsigned int,unsigned int,unsigned int,
			        unsigned int,unsigned int,unsigned int );
static void		timer_interrupt( int,void *,struct pt_regs * );


#if defined(CONFIG_ARCH_P2001)
static struct irqaction irqtimer2 = { timer_interrupt, SA_SHIRQ, 0, "timer",
				      NULL, NULL};
#else
#error XXX NOT SUPPORTED
#endif

/* ************************************************************************** */
/* *** localtime for linux kernel ******************************************* */

/* ************************************************************************** */
/* *** HT1381 Real Time Clock *********************************************** */
#ifdef INCLUDE_HW_RTC_HT1381
/* This is adapted from glibc */
/* Copyright (C) 1991, 1993 Free Software Foundation, Inc */

#define SECS_PER_HOUR 3600L
#define SECS_PER_DAY  86400L

/* Nonzero if YEAR is a leap year (every 4 years,
   except every 100th isn't, and every 400th is).  */
#define __isleap(year)  \
  ((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))

static const unsigned short int __mon_lengths[2][12] = {
	/* Normal years.  */
	{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
	/* Leap years.  */
	{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
};

struct tm {
	int	tm_sec;
	int	tm_min;
	int	tm_hour;
	int	tm_mday;
	int	tm_mon;
	int	tm_year;
	int	tm_wday;
	int	tm_yday;
};

static void
__tm_conv( struct tm *tmbuf,time_t *t,time_t offset )
{
	const unsigned short int	*ip;
	long				days;
	long				rem;
	int				y;

	days = *t / SECS_PER_DAY;
	rem = *t % SECS_PER_DAY;
	rem += offset;
	while (rem < 0) {
		rem += SECS_PER_DAY;
		--days;
	}
	while (rem >= SECS_PER_DAY) {
		rem -= SECS_PER_DAY;
		++days;
	}
	tmbuf->tm_hour = rem / SECS_PER_HOUR;
	rem %= SECS_PER_HOUR;
	tmbuf->tm_min = rem / 60;
	tmbuf->tm_sec = rem % 60;
	/* January 1, 1970 was a Thursday.  */
	tmbuf->tm_wday = (4 + days) % 7;
	if (tmbuf->tm_wday < 0)
		tmbuf->tm_wday += 7;
	y = 1970;
	while (days >= (rem = __isleap(y) ? 366 : 365)) {
		++y;
		days -= rem;
	}
	while (days < 0) {
		--y;
		days += __isleap(y) ? 366 : 365;
	}
	tmbuf->tm_year = y - 1900;
	tmbuf->tm_yday = days;
	ip = __mon_lengths[__isleap(y)];
	for (y = 0; days >= ip[y]; ++y)
		days -= ip[y];
	tmbuf->tm_mon = y;
	tmbuf->tm_mday = days + 1;
}

static unsigned char	rtc_buf[8];
static unsigned char	rtc_buf2[8];
static unsigned char	rtc_chk[8] = {0x81,0x25,0x0a,0x25,0x0a,0x05,0x05,0x00};  // Uhr anhalten
static unsigned char	rtc_msk[8] = {0xff,0x7f,0x1f,0x3f,0x1f,0x0f,0xff,0x00};
static int		rtc_valid = 0;

static void
rtc_wait( void )
{
	p2001_tm_t	*tcntl = (p2001_tm_t *) TIMER_BASE_ADDR;
	unsigned long	v1;
	unsigned long	v2;

	while (1) {
		v1 = tcntl->freerun_timer;
		while (1) {
			v2 = tcntl->freerun_timer;
			if (v1 > v2) {
				break;
			}
			if ((v2 - v1) > 0x20) {
				return;
			}
		}
	}
}

#define GPIO_OUT(x)	\
	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO_Out) = (x);	\
	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO_Out) = (x)

#define GPIO_IN(x)	\
	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO_IN) & (x)

#define GPIO2_OUT(x)	\
	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO2_Out) = (x);\
	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO2_Out) = (x)

#define GPIO2_IN(x)	\
	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO2_IN) & (x)

#if defined(PROJECT_T240) || defined(PROJECT_T444)
/*
 * RTC_CLK	= GPIO 26
 * RTC_IO	= GPIO 27
 * RTC_RESET	= LED_LATCH Bit 8
 */
#define RTC_RESET_OFF	davosv2_LedLatch(0x80, 0x80); rtc_wait()
#define RTC_RESET	davosv2_LedLatch(0x80, 0); rtc_wait()

#define RTC_CLK_LH	GPIO2_OUT( 0x4000400 ) 
#define RTC_CLK_HL	GPIO2_OUT( 0x4000000 ) 

#define RTC_IO_H	GPIO2_OUT( 0x8000800 ) 
#define RTC_IO_L	GPIO2_OUT( 0x8000000 ) 

#define RTC_IO_OUTPUT	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO2_En) |= 0x00000c00
#define RTC_IO_INPUT	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO2_En) &= 0xfffff7ff
#define RTC_IO_READ	GPIO2_IN( 0x00000800 )

#elif defined(PROJECT_EUMEX820) || defined(PROJECT_T484)
/*
 * RTC_CLK	= GPIO 4
 * RTC_IO	= GPIO 3
 * RTC_RESET	= GPIO 5
 */
#define RTC_RESET_OFF	GPIO_OUT( 0x00300020 ); rtc_wait()
#define RTC_RESET	GPIO_OUT( 0x00300000 ); rtc_wait()

#define RTC_CLK_LH	GPIO_OUT( 0x00100010 )
#define RTC_CLK_HL	GPIO_OUT( 0x00100000 )

#define RTC_IO_H	GPIO_OUT( 0x00080008 )
#define RTC_IO_L	GPIO_OUT( 0x00080000 )


#define RTC_IO_OUTPUT	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO_En) |= 0x00000008
#define RTC_IO_INPUT	*(volatile unsigned long *)(GPIO_BASE_ADDR+Adr_GPIO_En) &= 0xfffffff7
#define RTC_IO_READ	GPIO_IN( 0x00000008 )

#else
#error UNKNOWN PROJECT
#endif

#define RTC_CLK		rtc_wait(); RTC_CLK_LH; rtc_wait(); RTC_CLK_HL; rtc_wait()

static void
rtc_read( unsigned char *buf )
{
	int		i;
	int		j;
	unsigned char	val;

	/* release reset, clock low */
	RTC_RESET;

	RTC_CLK_HL;
	RTC_IO_OUTPUT;

	rtc_wait();
	RTC_RESET_OFF;
	/* i/o pin as output */

	rtc_wait();

	
	RTC_IO_H;	
	RTC_CLK;	// Bit 0: 1
	RTC_CLK;	// Bit 1: 1
	RTC_CLK;	// Bit 2: 1
	RTC_CLK;        // Bit 3: 1
	RTC_CLK;        // Bit 4: 1
	RTC_CLK;        // Bit 5: 1
	RTC_IO_L;	// 
	RTC_CLK;        // Bit 6: 0
	RTC_IO_H;	// 

	rtc_wait(); 
	RTC_CLK_LH;     // Bit 7: 1
	rtc_wait();

	/* i/o pin as input */
	RTC_IO_INPUT;
	RTC_IO_INPUT;

	for (i=0; i<8; i++) {
		for (j=0, val=0; j<8; j++) {
			RTC_CLK_HL; rtc_wait();
			if ( RTC_IO_READ  ) {
				val |= 1 << j;
			}
			RTC_CLK_LH; rtc_wait();
		}
		buf[i] = val;
	}

	/* reset, clock low */
	RTC_CLK_HL;
	RTC_RESET;
}

static void
rtc_write( unsigned char *buf )
{
	int		i;
	int		j;
	unsigned char	val;

	/* release reset, clock low */
	RTC_RESET_OFF;

	/* i/o pin as output */
	RTC_IO_OUTPUT;

	rtc_wait();

	RTC_IO_L;        
	RTC_CLK;	/* Bit 0: 0  R 1 W 0 */
	RTC_IO_H;
	RTC_CLK;	/* Bit 1: 1 */
	RTC_CLK;	/* Bit 2: 1 */
	RTC_CLK;	/* Bit 3: 1 */
	RTC_CLK;	/* Bit 4: 1 */
	RTC_CLK;	/* Bit 5: 1 */
	RTC_IO_L;
	RTC_CLK;        /* Bit 6: 0 */
	RTC_IO_H;
	RTC_CLK;        /* Bit 7: 1 */


	for (i=0; i<8; i++) {
		for (j=0, val=buf[i]; j<8; j++) {
			if (val & (1 << j)) {
				RTC_IO_H;
			} else {
				RTC_IO_L;
			}
			RTC_CLK;
		}
	}

	/* reset, clock low */
	RTC_RESET;
}

static void
rtc_writeprotect( int wp )
{
	int		i;
	unsigned char	val;

	/* release reset, clock low */
	RTC_RESET_OFF;

	/* i/o pin as output */
	RTC_IO_OUTPUT;

	rtc_wait();


	RTC_IO_L; 
	RTC_CLK;	/* Bit 0: 0 R1 W0 */
	RTC_IO_H; 
	RTC_CLK;	/* Bit 1: 1 A0 */
	RTC_CLK;        /* Bit 2: 1 A1 */
	RTC_CLK;	/* Bit 3: 1 A2 */
	RTC_IO_L; 
	RTC_CLK;	/* Bit 4: 0 */
	RTC_CLK;	/* Bit 5: 0 */
	RTC_CLK;	/* Bit 6: 0 */
	RTC_IO_H;
	RTC_CLK;	/* Bit 7: 1 */

	if (wp) {
		val = 0x80;
	} else {
		val = 0x00;
	}

	for (i=0; i<8; i++) {
		if (val & 1) {
			RTC_IO_H;
		} else {
			RTC_IO_L;
		}
		RTC_CLK;
		val = val >> 1;
	}

	/* reset, clock low */
	RTC_RESET;
}

static time_t
rtc_gettime( void )
{
	time_t		curr_time;
	time_t		diff_time;
	unsigned int	year;
	unsigned int	mon;
	unsigned int	day;
	unsigned int	hour;
	unsigned int	min;
	unsigned int	sec;
	
	diff_time = mktime( 2005,1,1,0,0,0 );
	memset( rtc_buf,0,sizeof( rtc_buf ));
	memset( rtc_buf2,0,sizeof( rtc_buf2 ));
	rtc_read( rtc_buf );

	if (!rtc_valid) {
		int		i;

		rtc_writeprotect( 0 );
		rtc_write( rtc_chk );
		rtc_writeprotect( 1 );
		rtc_read( rtc_buf2 );
		for (i=0; i<8; i++) {
			rtc_buf2[i] &= rtc_msk[i];
		}
		if (memcmp( rtc_chk,rtc_buf2,8 )) {
#if 1
			printk( KERN_NOTICE"[MSG] rtc: no realtime-clock present\n" );
#endif
			return( diff_time );
		}

		rtc_valid = 1;
		rtc_writeprotect( 0 );
	        rtc_buf[0] &=  0x7f;  // Clock Run
		rtc_write( rtc_buf );
		rtc_writeprotect( 1 );
	}

	/* Clock HALT & Clock WriteProtect => no battery for clock */
	if ((rtc_buf[0] & 0x80) && (rtc_buf[7] & 0x80)) {
		printk( KERN_NOTICE"[MSG] rtc: no battery, using default date\n" );
		return( diff_time );
	}
	rtc_buf[0] &= 0x7f;		/* sec */
	rtc_buf[1] &= 0x7f;		/* min */
	rtc_buf[2] &= 0x1f;		/* hour */
	rtc_buf[3] &= 0x3f;		/* day */
	rtc_buf[4] &= 0x1f;		/* month */
	rtc_buf[5] &= 0x0f;		/* day of week */

	sec	= BCD_TO_BIN( rtc_buf[0] );
	min	= BCD_TO_BIN( rtc_buf[1] );
	hour	= BCD_TO_BIN( rtc_buf[2] );
	day	= BCD_TO_BIN( rtc_buf[3] );
	mon	= BCD_TO_BIN( rtc_buf[4] );
	year	= BCD_TO_BIN( rtc_buf[6] ) + 2000;

#if 0
	printk( KERN_INFO"[INF] rtc: using y=%d m=%d d=%d h=%d m=%d s=%d\n",
		year,mon,day,hour,min,sec );
#endif

	curr_time = mktime( year,mon,day,hour,min,sec );

	if (curr_time < diff_time) {
		printk( KERN_WARNING"[WRN] rtc: has wrong date, using default date\n" );
		return( diff_time );
	} else {
		return( curr_time );
	}
}

static void
rtc_settime( time_t new_time )
{
	time_t		curr_time;
	struct tm	mytm;

	if (!rtc_valid) {
		return;
	}

	curr_time = rtc_gettime();
	if (new_time == curr_time) {
		return;
	}

	__tm_conv( &mytm,&new_time,0 );
	mytm.tm_mon += 1;
	mytm.tm_wday += 1;
	mytm.tm_year -= 100;

	rtc_buf[0] = BIN_TO_BCD( mytm.tm_sec ) & 0x7f;
	rtc_buf[1] = BIN_TO_BCD( mytm.tm_min ) & 0x7f;
	rtc_buf[2] = BIN_TO_BCD( mytm.tm_hour ) & 0x1f;
	rtc_buf[3] = BIN_TO_BCD( mytm.tm_mday ) & 0x3f;
	rtc_buf[4] = BIN_TO_BCD( mytm.tm_mon ) & 0x1f;
	rtc_buf[5] = BIN_TO_BCD( mytm.tm_wday ) & 0x0f;
	rtc_buf[6] = BIN_TO_BCD( mytm.tm_year );
	rtc_buf[7] = 0x00;

	rtc_writeprotect( 0 );
	rtc_write( rtc_buf );
	rtc_writeprotect( 1 );
}
#endif

/* Converts Gregorian date to seconds since 1970-01-01 00:00:00.
 * Assumes input in normal date format, i.e. 1980-12-31 23:59:59
 * => year=1980, mon=12, day=31, hour=23, min=59, sec=59.
 *
 * [For the Julian calendar (which was used in Russia before 1917,
 * Britain & colonies before 1752, anywhere else before 1582,
 * and is still in use by some communities) leave out the
 * -year/100+year/400 terms, and add 10.]
 *
 * This algorithm was first published by Gauss (I think).
 *
 * WARNING: this function will overflow on 2106-02-07 06:28:16 on
 * machines were long is 32-bit! (However, as time_t is signed, we
 * will already get problems at other places on 2038-01-19 03:14:08)
 */
static unsigned long
mktime( unsigned int year,
        unsigned int mon,
        unsigned int day,
        unsigned int hour,
        unsigned int min,
        unsigned int sec )
{
	if (0 >= (int) (mon -= 2)) {	/* 1..12 -> 11,12,1..10 */
		mon += 12;	/* Puts Feb last since it has leap day */
		year -= 1;
	}
	return((((unsigned long)(year/4 - year/100 + year/400 + 367*mon/12 + day) + year*365 - 719499)*24 + hour)*60 + min)*60 + sec;
}


void
do_gettimeofday( struct timeval *tv )
{
	unsigned long	flags;

	save_flags_cli( flags );
	*tv = xtime;
	tv->tv_usec += gettimeoffset();
	if (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
	restore_flags( flags );
}

void
do_settimeofday( struct timeval *tv )
{
	unsigned long	flags;

	save_flags_cli( flags );
#ifdef INCLUDE_HW_RTC_HT1381
	rtc_settime( tv->tv_sec );
#endif
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= gettimeoffset();

	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		/* stop active adjtime() */
	time_status |= STA_UNSYNC;
	time_state = TIME_ERROR;	/* p. 24, (a) */
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	restore_flags( flags );
}

/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick.
 */
static void
timer_interrupt( int irq,void *dev_id,struct pt_regs *regs )
{
	do_timer( regs );
	//update_rtc();
}

/*
 * Set up timer interrupt, and return the current time in seconds.
 */
static unsigned long
p2001_setup_timer( void )
{
	p2001_tm_t	*tcntl = (p2001_tm_t *) TIMER_BASE_ADDR;
	unsigned long	prediv;
	unsigned long	preload;

	/* 100 ticks per sec.  See also gettimeoffset above */
	prediv = 192;
	preload = SYSCLK/(prediv*100);
	tcntl->timer12_prediv |= TIMER_PREDIV_DISABLE_2;
	tcntl->timer_int &= ~TIMER_INT_TIMER2_EN;

	tcntl->timer12_prediv &= ~TIMER_PREDIV_PREDIV_2_MASK;
        /* special p200x timer feature prediv=192 has to be prediv -=1 */
	tcntl->timer12_prediv |= (prediv-1) << TIMER_PREDIV_PREDIV_2_SHIFT;
	tcntl->timer_preload &= ~TIMER_PRELOAD_2_MASK;
	tcntl->timer_preload |= preload << TIMER_PRELOAD_2_SHIFT;

	tcntl->timer_int |= TIMER_INT_TIMER2_EN;
	tcntl->timer12_prediv &= ~TIMER_PREDIV_DISABLE_2;
#ifdef INCLUDE_HW_RTC_HT1381
	return( rtc_gettime());
#else
	return( mktime(2004,1,1,0,0,0 ));
#endif
}

void time_init(void)
{
	xtime.tv_sec = p2001_setup_timer();
	xtime.tv_usec = 0;

	setup_arm_irq(IRQ_TC2, &irqtimer2);
}
