/*
 *	NET3	IP device support routines.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 *	Derived from the IP parts of dev.c 1.0.19
 * 		Authors:	Ross Biro, <bir7@leland.Stanford.Edu>
 *				Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *				Mark Evans, <evansmp@uhura.aston.ac.uk>
 *
 *	Additional Authors:
 *		Alan Cox, <gw4pts@gw4pts.ampr.org>
 */

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 *      
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/config.h>	/* For CONFIG_IP_CLASSLESS */
 
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/if_ether.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <net/ip.h>
#include <net/route.h>
#include <net/protocol.h>
#include <net/tcp.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <net/arp.h>

/* 
 *	Determine a default network mask, based on the IP address. 
 */
 
unsigned long ip_get_mask(unsigned long addr)
{
  	unsigned long dst;

  	if (addr == 0L) 
  		return(0L);	/* special case */

  	dst = ntohl(addr);
  	if (IN_CLASSA(dst)) 
  		return(htonl(IN_CLASSA_NET));
  	if (IN_CLASSB(dst)) 
  		return(htonl(IN_CLASSB_NET));
  	if (IN_CLASSC(dst)) 
  		return(htonl(IN_CLASSC_NET));
  
  	/*
  	 *	Something else, probably a multicast. 
  	 */
  	 
  	return(0);
}

/* 
 *	Check the address for our address, broadcasts, etc. 
 *
 *	I intend to fix this to at the very least cache the last
 *	resolved entry.
 */
 
int ip_chk_addr(unsigned long addr)
{
	struct device *dev;
#ifndef CONFIG_IP_CLASSLESS
	unsigned long mask;
#endif

#if 0
	printk( KERN_ERR">>> ip_chk_addr %s\n",in_ntoa( addr ) );
#endif

	/* 
	 *	Accept both `all ones' and `all zeros' as BROADCAST. 
	 *	(Support old BSD in other words). This old BSD 
	 *	support will go very soon as it messes other things
	 *	up.
	 *	Also accept `loopback broadcast' as BROADCAST.
	 */

	if (addr == INADDR_ANY || addr == INADDR_BROADCAST ||
	    addr == htonl(0x7FFFFFFFL))
	{
		return IS_BROADCAST;
	}

#ifndef  CONFIG_IP_CLASSLESS
	mask = ip_get_mask(addr);

	/*
	 *	Accept all of the `loopback' class A net. 
	 */
	 
	if ((addr & mask) == htonl(0x7F000000L)) {
		return IS_MYADDR;
	}
#else
	if ((addr & htonl(0x7F000000L)) == htonl(0x7F000000L)) {
		return IS_MYADDR;
	}
#endif

	/*
	 *	OK, now check the interface addresses. We could
	 *	speed this by keeping a dev and a dev_up chain.
	 */
	 
	for (dev = dev_base; dev != NULL; dev = dev->next) 
	{
		if ((!(dev->flags & IFF_UP)) || dev->family!=AF_INET)
			continue;
#if 1
#warning 22.04.2003 jse: dev->pa_addr == 0 removed (readme as comment)
/*
 * Fehlerbild: syslogd hat beim Internetverbindungsaufbau ploetzlich keine
 * Meldungen (nur via TCP) mehr ausgegeben.
 * Fehler: Flushen der FW bewirkt flushen des Arp-Cache. Das naechste write
 * zum TCP-Clienten loeste ein arp_find -> arp_set_predefined -> ip_chk_addr
 * mit der Adresse des TCP-Clienten 192.168.1.110 aus. ip_chk_addr hat nun
 * das ppp0 device mit der pa_addr == 0 gefunden und IS_MYADDR zurueckgegeben.
 * arp_find hat nun die MAC-Adresse von ssw0 fuer die Adresse 192.168.1.110
 * eingetragen und danach ging nichts mehr.
 */
#else
		/*
		 *	If the protocol address of the device is 0 this is
		 *	special and means we are address hunting (eg bootp).
		 */
		if (dev->pa_addr == 0) {
			return IS_MYADDR;
		}
#endif
		/*
		 *	Is it the exact IP address? 
		 */
		 
		if (addr == dev->pa_addr) {
			return IS_MYADDR;
		}
		/*
		 *	Is it our broadcast address? 
		 */
		 
		if ((dev->flags & IFF_BROADCAST) && addr == dev->pa_brdaddr) {
			return IS_BROADCAST;
		}
		/*
		 *	Nope. Check for a subnetwork broadcast. 
		 */
		 
		if (((addr ^ dev->pa_addr) & dev->pa_mask) == 0) 
		{
			if ((addr & ~dev->pa_mask) == 0) {
				return IS_BROADCAST;
			}
			if ((addr & ~dev->pa_mask) == ~dev->pa_mask) {
				return IS_BROADCAST;
			}
		}
		
#ifndef CONFIG_IP_CLASSLESS
		/*
	 	 *	Nope. Check for Network broadcast. 
	 	 */
	 	 
		if (((addr ^ dev->pa_addr) & mask) == 0) 
		{
			if ((addr & ~mask) == 0) {
				return IS_BROADCAST;
			}
			if ((addr & ~mask) == ~mask) {
				return IS_BROADCAST;
			}
		}
#endif
	}
	if(IN_MULTICAST(ntohl(addr))) {
		return IS_MULTICAST;
	}

	return 0;		/* no match at all */
}


/*
 *	Retrieve our own address.
 *
 *	Because the loopback address (127.0.0.1) is already recognized
 *	automatically, we can use the loopback interface's address as
 *	our "primary" interface.  This is the address used by IP et
 *	al when it doesn't know which address to use (i.e. it does not
 *	yet know from or to which interface to go...).
 */
 
unsigned long ip_my_addr(void)
{
  	struct device *dev;

  	for (dev = dev_base; dev != NULL; dev = dev->next) 
  	{
		if (dev->flags & IFF_LOOPBACK) 
			return(dev->pa_addr);
  	}
  	return(0);
}

/*
 *	Find an interface that can handle addresses for a certain address. 
 */

struct device * ip_dev_bynet(unsigned long addr, unsigned long mask)
{
	struct device *dev;
	struct device *best_dev = NULL;
	__u32  best_mask = mask;

	for (dev = dev_base; dev; dev = dev->next) 
	{
		if (!(dev->flags & IFF_UP))
			continue;
		if (dev->flags & IFF_POINTOPOINT)
		{
			if (addr == dev->pa_dstaddr)
				return dev;
			continue;
		}
		if (dev->pa_mask & (addr ^ dev->pa_addr))
			continue;
		if (mask == dev->pa_mask)
			return dev;
		if (best_dev && (best_mask & dev->pa_mask) != best_mask)
			continue;
		best_dev = dev;
		best_mask = dev->pa_mask;
	}
	return best_dev;
}

/*
 *	Find the first device with a given source address.
 */
 
struct device *ip_dev_find(unsigned long addr)
{
	struct device *dev;
	for(dev = dev_base; dev; dev=dev->next)
	{
		if((dev->flags&IFF_UP) && dev->pa_addr==addr)
			return dev;
	}
	return NULL;
}

struct device *dev_getbytype(unsigned short type)
{
	struct device *dev;

	for (dev = dev_base; dev != NULL; dev = dev->next) 
	{
		if (dev->type == type && !(dev->flags&(IFF_LOOPBACK|IFF_NOARP)))
			return(dev);
	}
	return(NULL);
}

