/*

IP_MASQ_NETMEETING Netmeeting masquerading module
Author:	Alex Nicolaou, LiquiMedia Inc., http://www.liquimedia.com/
        http://www.cgl.uwaterloo.ca/~anicolao/ to contact the author
	Copyright (C) 1999 LiquiMedia Inc. Distributed under GPL.

This module deals with allowing Netmeeting -> Netmeeting calls through
a firewall with IP Masquerade turned on. I have tested it with Netmeeting 3
calling Netmeeting 3 and with SunForum 3 calling Netmeeting 3. 

The reason this is necessary has to do with the fact that the protocols
netmeeting uses embed the destination IP address and port number inside
application level data. The module works by scanning through the packet
data and fiddling the insides in order to ensure that the packets get
properly redirected.

Here's what happens:

1. A machine within the firewall, which we'll call the "caller",  
initiates a netmeeting call to a machine outside the firewall, 
which we'll call the "callee". This is done by opening a TCP 
connection on port 1720.

2. A few packets later, the caller sends the callee a packet on the
existing 1720 connection which includes the caller's IP address. I
rewrite this to the firewall's address, although this may not be
necessary, strictly speaking. It is only after this packet is sent
that additional channels of communication are opened.

3. The callee sends to the caller a request to open a new tcp channel, 
using the existing 1720 channel. We'll refer to this new connection as
the secondary TCP channel. The request to open the channel can be 
identified because the application level data includes the callee's 
IP address and the desired port for the connection.

4. The caller opens the secondary TCP channel.

5. The caller sends to the callee open channel requests that encode
the caller's IP and a port number; these open channel requests 
correspond to channels for UDP data carrying audio and video packets.
The module adds entries to the masq tables for these channels and 
rewrites the data going to the callee to point at the firewall's 
address and the masq'ed port number.

6. Audio and video data arrives at the firewall on the masq'ed port
and the packets are redirected to the appropriate caller.

7. The module limits the number of structures it is willing to allocate
so that it won't use up all of the kernel memory in the event of many
converstaions through the firewall. The number of monitored ports is
limited by monitor_limit, which can be set at insmod time with a
command like "insmod ip_masq_netmeeting.o monitor_limit=40". The default
limit is 16, which will allow several simultaneous netmeeting connections. 

For incoming calls, you need to manually forward several ports by
hand to a single machine inside your firewall. On 2.0.X kernels, 
the commands you need to enter are something like:

ipautofw -F
ipautofw -A -r tcp 1503 1503 -h 10.0.0.1
ipautofw -A -r tcp 1720 1720 -h 10.0.0.1
ipautofw -A -r tcp 1731 1731 -h 10.0.0.1

Note that the first one flushes all forwarding rules, so if you are
using forwarding for something else you'll probably not want to do 
that part. After that, incoming calls should ring on the given IP
(10.0.0.1 in this example) and you can answer and talk as usual. Even 
in the incoming call case, the machine inside your firewall is still
referred to as the "caller". 

License:

	This program is free software; you can redistribute it and/or
 	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
 	2 of the License, or (at your option) any later version.
*/

/*
 * Modifications for P2001 Copyright (C) <2001-2008> <Funkwerk Enterprise Communications GmbH>
 * Written by Juergen Selent (juergen.selent@funkwerk-ec.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License. 
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS * FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>. 
 *      
 *
 * DISCLAIMER OF WARRANTY 
 *
 * THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY 
 * APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE 
 * COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" 
 * WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, 
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY 
 * AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM 
 * PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, 
 * REPAIR OR CORRECTION. 
 *  
 * LIMITATION OF LIABILITY 
 *
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR 
 * CONVEYS THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF 
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
 * HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES. 
 */

#include <linux/module.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <net/protocol.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <net/ip_masq.h>

/*
 * Debug macros
 */
#define DEB(fmt, args...) \
	printk( KERN_DEBUG/**/"ip_masq_h323: [DEB] "/**/fmt, ##args );
#define INF(fmt, args...) \
	printk( KERN_NOTICE/**/"ip_masq_h323: [INF] "/**/fmt, ##args );
#define WRN(fmt, args...) \
	printk( KERN_WARNING/**/"ip_masq_h323: [WRN] "/**/fmt, ##args );
#define ERR(fmt, args...) \
	printk( KERN_ERR/**/"ip_masq_h323: [ERR] "/**/fmt, ##args );


extern char *in_ntoa(unsigned long);

/* limit the number of forwarding entries to this, settable as a parameter */
#define MONITOR_LIMIT	16

#if   (LINUX_VERSION_CODE >= 0x020400)
#error this module has never been tested/compiled with Linux 2.4

#elif (LINUX_VERSION_CODE >= 0x020300)
#error this module has never been tested/compiled with Linux 2.3

#elif (LINUX_VERSION_CODE >= 0x020200)

#define GET_IP_HEADER(skb) (skb->nh.iph)

#define IP_MASQ_NEW_UDP(saddr, port, daddr) \
	ip_masq_new(IPPROTO_UDP, ms->maddr, 0, saddr, port, daddr, \
		    0, IP_MASQ_F_NO_DPORT)

#define IP_MASQ_GET_UDP(saddr, port, daddr) \
	ip_masq_out_get(IPPROTO_UDP, saddr, port, daddr, 0)

#define IP_MASQ_NEW_TCP(saddr, port, daddr) \
	ip_masq_new(IPPROTO_TCP, ms->maddr, 0, saddr, port, daddr, \
		    0, IP_MASQ_F_NO_DPORT)

#define IP_MASQ_GET_TCP(saddr, port, daddr) \
	ip_masq_out_get(IPPROTO_TCP, saddr, port, daddr, 0)

#define IP_MASQ_KEEPALIVE(n_ms) \
	{/* no-op in v2.2, done by GET_UDP */}

#define DEV_T __u32

#elif (LINUX_VERSION_CODE >= 0x020100)
#error this module has never been tested/compiled with Linux 2.1

#elif (LINUX_VERSION_CODE >= 0x020000)
#define GET_IP_HEADER(skb) (skb->h.iph)

#define IP_MASQ_NEW_UDP(saddr, port, daddr) \
	ip_masq_new(dev, IPPROTO_UDP, saddr, port, daddr, \
		    0, IP_MASQ_F_NO_DPORT)

#define IP_MASQ_GET_UDP(saddr, port, daddr) \
	ip_masq_out_get_2(IPPROTO_UDP, saddr, port, daddr, 0)

#define IP_MASQ_NEW_TCP(saddr, port, daddr) \
	ip_masq_new(dev, IPPROTO_TCP, saddr, port, daddr, \
		    0, IP_MASQ_F_NO_DPORT)

#define IP_MASQ_GET_TCP(saddr, port, daddr) \
	ip_masq_out_get_2(IPPROTO_TCP, saddr, port, daddr, 0)

#define IP_MASQ_KEEPALIVE(n_ms) \
	ip_masq_set_expire(n_ms,0)

#define EXPORT_NO_SYMBOLS \
        register_symtab(0)

#define DEV_T struct device*

#else
#error Your version of Linux is too old. How about upgrading?
#endif

/* singly linked list of ports currently being watched */
typedef struct _port_list_entry {
	struct _port_list_entry *next;
	int port;
	struct ip_masq_app masq;
} PortListEntry;

static PortListEntry *ports = NULL;
static PortListEntry *unloadqueue = NULL;

/*
   adds a new portnumber to the list of ports being monitored by this code 
*/
static int ip_masq_netmeeting_monitor_port(int portno);
static int ip_masq_netmeeting_unqueue_port(int portno);
static int ip_masq_netmeeting_unregister_list(PortListEntry **cptr);

static int
masq_netmeeting_init_1 (struct ip_masq_app *mapp, struct ip_masq *ms)
{
	DEB( "masq_netmeeting_init_1: ENTER, port %u\n",
	     (unsigned short)(mapp->type&0xffff) );
	DEB( "   caller: %s:%d\n", 
		in_ntoa( ms->saddr ),ntohs( ms->sport ));
	DEB( "   firewall: %s:%d\n", 
		in_ntoa( ms->maddr ),ntohs( ms->mport ));
	DEB( "   callee: %s:%d\n",
		in_ntoa( ms->daddr ),ntohs( ms->dport ));
        MOD_INC_USE_COUNT;
        return( 0 );
}

static int
masq_netmeeting_done_1 (struct ip_masq_app *mapp, struct ip_masq *ms)
{
	DEB( "masq_netmeeting_done_1: ENTER\n" );
	if (ntohs(ms->sport) != 1720) {
		DEB( "putting %d on unload queue (mapp->n_attach=%d)\n",
		     ntohs( ms->sport ),mapp->n_attach );
		ip_masq_netmeeting_unqueue_port( ntohs( ms->sport ));
	}
        MOD_DEC_USE_COUNT;
        return( 0 );
}

#define ARRAYSIZE(X) (sizeof(X)/sizeof(X[0]))

#define UNUSED				1
#define OPENED				2	/* conversation has started but channel opens aren't expected */
#define NEXT_CHANNEL_MATTERS		3	/* next channel open needs to be monitored */
#define DONE				4	/* not yet used */
#define SECONDARY_TCP_CONNECTION	5	/* this state needs constant monitoring */

/*
   Both incoming and outgoing packets are processed by this filter, which
   does all necessary rewrites of addresses and adds entries to the masq
   table. This function is basically the entire module
*/
int
masq_netmeeting_io( struct ip_masq_app *mapp,
                    struct ip_masq *ms,
                    struct sk_buff **skb_p,
                    DEV_T dev )
{
	struct sk_buff *skb = *skb_p;
	struct iphdr *iph = GET_IP_HEADER(skb);
#warning XXX todo calc tcp header size
	struct tcphdr *th = (struct tcphdr *)&(((char *)iph)[iph->ihl*4]);
	unsigned char *data = (char *)&th[1];
	unsigned char *data_limit = skb->h.raw + skb->len;

	DEB( "masq_netmeeting_io: ENTER (skb->len=%ld)\n",skb->len );

	/* avoid overrun in loop below */
	data_limit -= 3;

	ip_masq_netmeeting_unregister_list( &unloadqueue );

	if (th->ack == 0) {
		if ((int)ms->app_data == 0) {
			(int)ms->app_data = UNUSED;
			DEB( "new connection\n" );
		}
		if (ntohs(th->dest) != 1720) {
			(int)ms->app_data = SECONDARY_TCP_CONNECTION;
			DEB( "monitoring second TCP on port %d->%d\n",
			     ntohs( th->source ),ntohs( th->dest ));
		} else {
			DEB( "primary netmeeting conn\n" );
			DEB( "   caller: %s:%d\n", 
				in_ntoa( ms->saddr ),ntohs( ms->sport ));
			DEB( "   firewall: %s:%d\n", 
				in_ntoa( ms->maddr ),ntohs( ms->mport ));
			DEB( "   callee: %s:%d\n",
				in_ntoa( ms->daddr ),ntohs( ms->dport ));
		}
	}
	if (th->fin == 1 && ntohs( ms->sport ) != 1720) {
		DEB( "fin -- should shortly unload support on %d\n",
		     ntohs( ms->sport ));
	}

	DEB( "analyzing %d->%d packet\n",ntohs( th->source ),ntohs( th->dest ));
	while (data < data_limit) {
		if ((int)ms->app_data == NEXT_CHANNEL_MATTERS &&
		    !memcmp( data,&(ms->daddr),4 ))
		{
			int ret;
			int port = (((int)(*(data+4)))<<8) | (*(data+5)); 

			DEB( "Callee's IP Address: %d.%d.%d.%d\n", 
			     data[0],data[1],data[2],data[3] );
			DEB( "channel from in->out opened on 0x%04x\n",port );

			ret = ip_masq_netmeeting_monitor_port( port );
			/* (int)ms->app_data = DONE; */

			DEB( "monitor ret: %d\n",ret );
		}

		if ((int)ms->app_data == OPENED &&
		    !memcmp( data,&(ms->daddr),4 ))
		{
			DEB( "next channel open matters\n" );
			(int)ms->app_data = NEXT_CHANNEL_MATTERS;
		} else if ((int)ms->app_data == OPENED && 
		           !memcmp( data,&(ms->maddr),4 ))
		{
			DEB( "someone is calling in; next ch open matters; rewrite\n" );
			(int)ms->app_data = NEXT_CHANNEL_MATTERS;
			memcpy( data,&(ms->saddr),4 );
			data++;
			continue;
		}

		if (!memcmp( data,&(ms->saddr),4 )) {
			/* our IP address */
			int port;

			DEB( "Caller's IP Address: %d.%d.%d.%d state=%d\n", 
			     data[0],data[1],data[2],data[3],(int)ms->app_data );
			DEB( "rewriting IP\n" );

			memcpy( data,&(ms->maddr),4 );
			data += 4;

			port = (((int)(*data)) << 8) | ((int)(*(data + 1)));
			if ((int)ms->app_data == SECONDARY_TCP_CONNECTION && port != 1503) {
				struct ip_masq	*n_ms;

				n_ms = IP_MASQ_GET_UDP( ms->saddr,
				                        htons( port ),
				                        ms->daddr );
				if (n_ms) {
					/* reset expiry to keep entry alive */
					IP_MASQ_KEEPALIVE( n_ms );
				} else {
					/* no entry yet; make one */
					n_ms = IP_MASQ_NEW_UDP( ms->saddr, 
					                        htons( port ),
					                        ms->daddr );
				}

				/* n_ms->mport contains the port on which data
				   must arrive to be correctly forwarded */
				memcpy( data,&(n_ms->mport),2 );

				DEB( "REWROTE port %d->%d\n",port,
				     (((int)(*data))<<8)|((int)(*(data+1))) );
				DEB( "forwarding UDP data\n" );
			} else if (port != 1503) {
				int ret;

				struct ip_masq	*n_ms;

				n_ms = IP_MASQ_GET_TCP( ms->saddr,
				                        htons( port ),
				                        ms->daddr );
				ret = ip_masq_netmeeting_monitor_port( port );
				DEB( "monitoring ret: %d\n",ret );

				if (n_ms) {
					/* reset expiry to keep entry alive */
					IP_MASQ_KEEPALIVE( n_ms );
				} else {
					/* no entry yet; make one */
					n_ms = IP_MASQ_NEW_TCP( ms->saddr, 
					                        htons( port ),
					                        ms->daddr );
				}

				/* n_ms->mport contains the port on which data
				   must arrive to be correctly forwarded */
				memcpy( data,&(n_ms->mport),2 );

				DEB( "REWROTE port %d->%d\n",port,
				     (((int)(*data))<<8)|((int)(*(data+1))) );
				DEB( "forwarding TCP data\n" );
			}
		}
		data++;
	}
	DEB( "masq_netmeeting_io: LEAVE\n" );
	return( 0 );
}

struct ip_masq_app ip_masq_netmeeting = {
        NULL,			/* next */
        "netmeeting",
        0,                      /* type */
        0,                      /* n_attach */
        masq_netmeeting_init_1,	/* ip_masq_init_1 */
        masq_netmeeting_done_1,	/* ip_masq_done_1 */
        masq_netmeeting_io,	/* pkt_out */
        masq_netmeeting_io    	/* pkt_in */
};


/*
 * 	ip_masq_netmeeting initialization
 */


static int
ip_masq_netmeeting_monitor_port( int portno )
{
	PortListEntry	*newmasq;
	PortListEntry	*counter;
	unsigned long	flags;
	int		limit_check;
	int		ret;

	DEB( "ip_masq_netmeeting_monitor_port: ENTER(%d)\n",portno );

	counter = ports;
	limit_check = MONITOR_LIMIT;

	while (counter && (counter = counter->next)) {
		limit_check--;
	}
	if (limit_check <= 0) {
		INF( "port limit exceeded\n" );
		return( -ENOMEM );
	}

	if ((newmasq = kmalloc( sizeof(*newmasq),GFP_ATOMIC )) == NULL) {
		ERR( "out of memory\n" );
		return( -ENOMEM );
	}


	memcpy( &(newmasq->masq),&ip_masq_netmeeting,sizeof( newmasq->masq ));
	if ((ret = register_ip_masq_app( &(newmasq->masq),IPPROTO_TCP,portno ))) {
		kfree( newmasq );
		return( ret );
	}
	DEB( "monitoring port %d\n",portno );

	newmasq->port = portno;
	save_flags_cli( flags );
	newmasq->next = ports;
	ports = newmasq;
	restore_flags( flags );

	return( 0 );
}

static int
ip_masq_netmeeting_unqueue_port( int portno )
{
	PortListEntry	*predecessor;
	PortListEntry	*removeme;
	unsigned long	flags;

	DEB( "ip_masq_netmeeting_unqueue_port: ENTER (%d)\n",portno );

	predecessor = NULL;
	save_flags_cli( flags );
	removeme = ports;

	while (removeme != NULL && removeme->port != portno) {
		predecessor = removeme;
		removeme = removeme->next;
	}

	if (removeme == NULL) {
		/* no matching port found? bail out */
		restore_flags( flags );
		WRN( "port %d not being monitored?\n",portno );
		return( -1 );
	}

	if (predecessor == NULL) {
		if (removeme != ports) {
			restore_flags( flags );
			WRN( "impossible?! (1)\n" );
			return( -1 );
		}
		ports = ports->next;
	} else {
		if (removeme != predecessor->next) {
			restore_flags( flags );
			WRN( "impossible?! (2)\n" );
			return( -1 );
		}
		predecessor->next = removeme->next;
	}

	removeme->next = unloadqueue;
	unloadqueue = removeme;
	restore_flags( flags );
	DEB( "queued remove on %d\n",removeme->port );

	return( 0 );
}

int
ip_masq_netmeeting_init( void )
{
	DEB( "ip_masq_netmeeting_init: ENTER\n" );
	return( ip_masq_netmeeting_monitor_port( 1720 ));
}

/*
 * 	ip_masq_netmeeting fin.
 */

static int
ip_masq_netmeeting_unregister_list( PortListEntry **cptr )
{
	PortListEntry	*freeme;
	PortListEntry	*curr;
	unsigned long	flags;
	int		tmp;
	int		ret;

	DEB( "ip_masq_netmeeting_unregister_list: ENTER\n" );

	save_flags_cli( flags );
	curr = *cptr;
	*cptr = NULL;
	restore_flags( flags );

	ret = 0;
	while (curr != NULL) {
		DEB( "unloading support on port %d, n_attach %d\n",
		     curr->port,curr->masq.n_attach );
		if ((tmp = unregister_ip_masq_app( &(curr->masq) ))) {
			ret = tmp;
			WRN( "unloading support on port %d failed, ret=%d\n",
			     curr->port,ret );
		}
		freeme = curr;
		curr = curr->next;
		kfree( freeme );
	}
	return( ret );
}

int
ip_masq_netmeeting_done( void )
{
	int	ret;

	DEB( "ip_masq_netmeeting_done: ENTER\n" );
	ret = ip_masq_netmeeting_unregister_list( &ports );
	ports = NULL;
	return( ret );
}

#ifdef MODULE
int
init_module( void )
{
        if (ip_masq_netmeeting_init() != 0) {
                return( -EIO );
	}
	EXPORT_NO_SYMBOLS;
        return( 0 );
}

void
cleanup_module( void )
{
        if (ip_masq_netmeeting_done() != 0) {
                printk( KERN_INFO"IP_MASQ_NETMEETING: can't remove module\n" );
	}
}
#endif /* MODULE */
